"""
bioBakery Workflows: utilities module
Utility functions for workflows and tasks

Copyright (c) 2016 Harvard School of Public Health

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import os
import sys
import math
import functools
import time
import collections

from anadama2.tracked import TrackedDirectory

# try to import urllib.request.urlretrieve for python3
try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve

def get_package_file(basename, type="template"):
    """ Get the full path to a file included in the installed python package.

        Args:
            basename (string) : The basename of the file
            type (string) : The type of file to find (template or Rscript)

        Returns: 
            string : The full path to the file

    """

    if type == "template":
        subfolder = "document_templates"
        extension = ".template.py"
    else:
        subfolder = "Rscripts"
        extension = ".R"

    # get all of the templates in this folder
    package_install_folder=os.path.join(os.path.dirname(os.path.realpath(__file__)), subfolder)
    found_files=filter(lambda file: file.endswith(extension),os.listdir(package_install_folder))

    # return the template with the name
    matching_file=list(filter(lambda file: file.startswith(basename+extension), found_files))

    if matching_file:
        matching_file=os.path.join(package_install_folder,matching_file[0])
    else:
        matching_file=""

    return matching_file
    


def change_pweave_figure_size_heatmap(pdf_format):
    """ Change the figure size for heatmaps based on the output format"""
    fig_size = (4,4) if pdf_format else (2.5,2.5)
    change_pweave_figure_size(fig_size)
    
def reset_pweave_figure_size():
    """ Set the pweave figure size back to the default """
    change_pweave_figure_size((8,6))
    
def change_pweave_figure_size(fig_size):
    """ Change the pweave default figure size """
    import pweave
    pweave.rcParams["chunk"]["defaultoptions"].update({'f_size': fig_size})

def byte_to_megabyte(byte):
    """
    Convert byte value to megabyte
    """
    
    return byte / (1024.0**2)

class ReportHook():
    def __init__(self):
        self.start_time=time.time()
        
    def report(self, blocknum, block_size, total_size):
        """
        Print download progress message
        """
        
        if blocknum == 0:
            self.start_time=time.time()
            if total_size > 0:
                print("Downloading file of size: " + "{:.2f}".format(byte_to_megabyte(total_size)) + " MB\n")
        else:
            total_downloaded=blocknum*block_size
            status = "{:3.2f} MB ".format(byte_to_megabyte(total_downloaded))
                    
            if total_size > 0:
                percent_downloaded=total_downloaded * 100.0 / total_size
                # use carriage return plus sys.stdout to overwrite stdout
                try:
                    download_rate=total_downloaded/(time.time()-self.start_time)
                    estimated_time=(total_size-total_downloaded)/download_rate
                except ZeroDivisionError:
                    download_rate=0
                    estimated_time=0
                estimated_minutes=int(estimated_time/60.0)
                estimated_seconds=estimated_time-estimated_minutes*60.0
                status +="{:3.2f}".format(percent_downloaded) + " %  " + \
                    "{:5.2f}".format(byte_to_megabyte(download_rate)) + " MB/sec " + \
                    "{:2.0f}".format(estimated_minutes) + " min " + \
                    "{:2.0f}".format(estimated_seconds) + " sec "
            status+="        \r"
            sys.stdout.write(status)

def read_file_n_lines(file,n):
    """ Read a file n lines at a time """

    line_set=[]
    with open(file) as file_handle:
        for line in file_handle:
            if len(line_set) == n:
                yield line_set
                line_set=[]
            line_set.append(line)

    # yield the last set
    if len(line_set) == n:
        yield line_set
            
def read_file_catch(file, delimiter="\t"):
    """ Try to read the file, catch on error. Split data by delimiter. """
    
    try:
        handle=open(file)
        lines=handle.readlines()
        handle.close()
    except EnvironmentError:
        sys.exit("Error: Unable to read file: "+file)
    
    # split rows by delimiter
    new_lines=[line.rstrip().split(delimiter) for line in lines]    
    
    return new_lines
            
def read_metadata(metadata_file, taxonomy_file, name_addition="", ignore_features=[], otu_table=False):
    """ Read in the metadata file. Samples can be rows or columns.
    Ignore features if set. 
        
    Args:
        metadata_file (string): The path to the metadata file.
        taxonomy_file (string): The path to a taxonomy file (or any file with
            the sample names as the column names).
        name_addition (string): Any strings that should be removed from the
            names in the taxonomy files to get the sample names.
        ignore_features (list): A list of strings of features to ignore
            when reading the metadata file.
        otu_table (bool): Set if otu table is used for taxonomy to allow for different
            header format.
        
    Returns:
        (list): A list of lists of the metadata (samples as columns)
    """
    
    # read in a taxonomy file to get the sample names from the columns
    if otu_table:
        samples=set([name.replace(name_addition,"") for name in read_file_catch(taxonomy_file)[0][1:-1]])
    else:
        samples=set([name.replace(name_addition,"") for name in read_file_catch(taxonomy_file)[0][1:]])
        
    # read in the metadata file
    data=read_file_catch(metadata_file)
    # check if the columns or rows are samples
    possible_samples=data[0][1:]
    overlap=samples.intersection(possible_samples)
    if len(list(overlap)) == 0:
        # the samples must be the rows so invert the data
        data=[list(a) for a in zip(*data)]
        possible_samples=data[0][1:]
        overlap=samples.intersection(possible_samples)
    
    # check for samples not included in metadata
    if len(list(overlap)) < len(list(samples)):
        sys.exit("ERROR: Not all of the samples in the data set have"+
            " metadata. Please review the metadata file. The following samples"+
            " were not found: "+",".join(list(samples.difference(possible_samples))))

    # remove any features that should be ignored
    new_data=[]
    for row in data:
        if not row[0] in ignore_features:
            new_data.append(row)
        else:
            ignore_features.remove(row[0])

    # check for any features that do not vary
    check_diff = {}
    for row in new_data[1:]:
        for value in row[1:]:
            if not row[0] in check_diff:
                check_diff[row[0]]=[]
            check_diff[row[0]]=check_diff[row[0]]+[value]

    for feature_name, feature_types in check_diff.items():
        if len(list(set(feature_types))) == 1:
            sys.exit("ERROR: Please remove feature '"+feature_name+"' as it"+
                " only includes a single type of '"+feature_types[0]+"'.")

    # check for any features that were not found
    if ignore_features:
        sys.exit("ERROR: Unable to find features that should be ignored: "+
            ",".join(ignore_features))
        
    return new_data

def label_metadata(data, categorical=[], continuous=[]):
    """ Label the metadata type. All numerical is continous.
    
    Args:
        data (lists of lists): A list of metadata 
        categorical (list): A list of categorical features.
        continuous (list): A list of continuous features.
        
    Returns:
        (dict): A dictionary of metadata labels
        (list): A list of lists of the metadata (converted to floats if continuous)
    """
    
    # add labels to the metadata, ignore sample names, convert to nan misisng values if continuous
    missing = ["Unknown", "unknown", "NA", "na", "nan", "NaN", "NAN", " "]
    labeled_data=[data[0]]
    labels={}
    for row in data[1:]:
        # check if there are missing values in row, convert to nan
        for index, item in enumerate(row):
            if item in missing or item == "":
                row[index] = "nan"
        # apply specific labels if set
        label=None
        if row[0] in continuous:
            try:
                row[1:] = map(float, row[1:])
                label="con"
            except ValueError:
                label="cat"
            continuous.remove(row[0])
        if row[0] in categorical:
            label="cat"
            categorical.remove(row[0])
        if not label:
            try:
                row[1:] = map(float, row[1:])
                label="con"
            except ValueError:
                label="cat"
        if label == "cat":
            # if label is categorical convert misisng values to Unknown
            for index, item in enumerate(row):
                if item == "nan":
                    row[index] = "NA"
        labeled_data.append(row)
        labels[row[0]]=label
        
    # check for remaining labels
    if categorical:
        sys.exit("ERROR: Unable to find and label categorical feature in metadata: "+
            ",".join(categorical))
    if continuous:
        sys.exit("ERROR: Unable to find and label continuous feature in metadata: "+
            ",".join(continuous))
        
    return labels, labeled_data

def filter_metadata_categorical(metadata, metadata_labels):
    """ Return only the metadata that is categorical 
    
    Args:
        metadata (lists of lists): A list of metadata.
        metadata_labels (dict): The labels for the metadata.
        
    Returns:
        metadata (lists of lists): Only the categorical metadata.
    """
    
    new_metadata=[]
    for row in metadata:
        if metadata_labels.get(row[0],"con") == "cat":
            new_metadata.append(row)
            
    return new_metadata

def group_samples_by_metadata(metadata, data, samples):
    """ Return the samples grouped by the metadata. The data and metadata
    should have the same ordering of sample columns.
    
    Args:
        metadata (list): A single metadata list.
        data (list): A single data set.
        samples (list): The samples (corresponding to the columns in the 
            data/metadata).
    
    Return:
        data (list of lists): The data organized by the metadata groups.
    """
    
    # get the samples for each metadata type
    sorted_samples_grouped={}
    for name, type in zip(samples,metadata[1:]):
        sorted_samples_grouped[type]=sorted_samples_grouped.get(type,[])+[name]
        
    sorted_data_grouped={}
    for row in data:
        sorted_temp={}
        for data_point, type in zip(row, metadata[1:]):
            sorted_temp[type]=sorted_temp.get(type,[])+[data_point]
            
        for key, value in sorted_temp.items():
            if not key in sorted_data_grouped:
                sorted_data_grouped[key]=[]
            sorted_data_grouped[key].append(value)
    
    return sorted_data_grouped, sorted_samples_grouped
    
def merge_metadata(metadata, samples, values, values_without_names=None):
    """ Merge the metadata and values into a single set. Samples are columns. 

    Args:
        metadata (lists of lists): A list of metadata. 
        samples (list): A list of samples that correspond with value columns.
        values (lists of lists): A list of values to merge with the metadata.
        values_without_names (bool): Set if the values do not have row names.
    Returns:
        (list): A list of lists of the merged data.
        (list): A list of the samples (might be a subset based on metadata available).
    """
    
    # get the indexes for the samples in the data that match with the metadata
    sample_index=[]
    metadata_index=[]
    samples_found=[]
    for index, name in enumerate(metadata[0][1:]):
        if name in samples:
            samples_found.append(name)
            sample_index.append(samples.index(name))
            metadata_index.append(index)
            
    # warn if no metadata samples match the values samples
    if len(sample_index) == 0:
        print("Warning: Metadata does not match samples.")
        return values, samples
    
    if len(samples) > len(metadata[0][1:]):
        print("Warning: Metadata only provided for a subset of samples.")
    
    # add metadata to the new data
    new_data=[]
    for row in metadata[1:]:
        # add only matching samples
        new_data.append([row[0]]+[row[i+1] for i in metadata_index])
        
    # add abundance values to the new data
    for row in values:
        if values_without_names:
            new_data.append([row[i] for i in sample_index])
        else:
            new_data.append([row[0]]+[row[i+1] for i in sample_index])
        
    return new_data, samples_found

def download_file(url, download_file):
    """
    Download a file from a url
    Create folder for downloaded file if it does not exist
    """

    create_folders(os.path.dirname(download_file))

    try:
        print("Downloading "+url)
        file, headers = urlretrieve(url,download_file,reporthook=ReportHook().report)
        # print final return to start new line of stdout
        print("\n")
    except EnvironmentError:
        print("WARNING: Unable to download "+url)

def try_log10(value):
    """ Try to convert value to log10 """
    
    try:
        new_value = math.log10(value)
    except ValueError:
        new_value = 0
        
    return new_value

def name_task(sample,software):
    """ Name the task based on the sample name and software """
    
    return software+"____"+os.path.basename(sample)

def add_to_list(items,new_item):
    """ Add the value to the list/tuple. If the item is not a list, create a new
        list from the item and the value 
        
    Args:
        items (list, string or tuple): Single or multiple items
        new_item (string): The new value
        
    Returns:
        (list): A list of all values
    """
    
    if isinstance(items,tuple):
        items=[i for i in items]
    
    if not isinstance(items,list):
        items=[items]
        
    return items+[new_item]

def metacyc_url(pathway):
    """ Return the url for the pathway on the MetaCyc website 
    
    Args:
        pathway (string): The MetaCyc pathway
        
    Returns
       (string): The url to the website for the pathway
       
    """
    
    return "http://metacyc.org/META/NEW-IMAGE?type=NIL&object="+pathway

def run_task(command, **keywords):
    """ Run the task command, formatting command with keywords. The command stdout
        and stderr are written to the workflow log.
    
    Args:
        command (string): A string to execute on the command line. It can be
            formatted the same as a task command.
       
    Returns:
        (int): Return code from command.     
    """

    from anadama2.helpers import format_command
    from anadama2.helpers import sh
    
    # format the command to include the items for this task
    command=format_command(command, **keywords)
    
    # run the command
    return_code = sh(command)()
    
    return return_code

def partial_function(function, **keywords):
    """ Return a partial function, setting function name attribute
    
    Args:
        function (function): A function
        keywords: One or more keywords to be applied to the function
        
    Returns:
        (function): A partial function
        
    """
    
    partial = functools.partial(function, **keywords)
    partial.__name__ = function.__name__
    
    return partial
        

def paired_files(files, extension, pair_identifier=None):
    """ Find sets of paired-end reads
    
    This function will find sets of paired end reads from a list of files.
    
    Args:
        files (list): A list of files (with or without the full paths)
        extension (string): The extension for all files.
        pair_identifier (string): The string in the file basename to identify
            the first pair in the set (optional).
            
    Requires:
        None
        
    Returns:
        list: A list of paired files.
        
    Example:
        paired_set = paired_reads(["1.R1.fq", "1.R2.fq"],".fq")
        
    """
    
    # add period to extension if not included
    if not extension.startswith("."):
        extension="."+extension
    
    if pair_identifier is None:
        pair_identifier=".R1"
    
    # check for the one in the pair identifier
    if not "1" in pair_identifier:
        sys.exit("Please provide the identifier for the first pair set (ie R1).")
    
    pair_identifier2=pair_identifier.replace("1","2",1)

    input_pair1 = list(filter(lambda file: os.path.basename(file).replace(extension,"").endswith(pair_identifier), files))
    input_pair2 = list(filter(lambda file: os.path.basename(file).replace(extension,"").endswith(pair_identifier2), files))
    
    # only return matching pairs of files in the same order
    paired_file_set = [[],[]]
    for file1 in sorted(input_pair1):
        # find the matching file in the second set
        name1=sample_names(file1, extension, pair_identifier)
        for file2 in input_pair2:
            name2=sample_names(file2, extension, pair_identifier2)
            if name1 and name1 == name2:
                paired_file_set[0].append(file1)
                paired_file_set[1].append(file2)
                input_pair2.remove(file2)
                break
    
    return paired_file_set

def sample_names(files,extension,pair_identifier=None):
    """ Return the basenames of the files, without any extensions, as the sample names
    
    Args:
        files (list): A list of files (with or without the full paths)
        extension (string): The extension for all files.
        pair_identifier (string): The string in the file basename to identify
            the first pair in the set (optional).
        
    Requires:
        None
        
    Returns:
        list: A list of sample names (file basenames)

    Example:
        names = sample_names(["1.R1.fq", "1.R2.fq"],".fq")
        
    """
    
    # add period to extension if not included
    if not extension.startswith("."):
        extension="."+extension
    
    # if files is a string, convert to a list
    convert=False
    if isinstance(files,str):
        files=[files]
        convert=True
        
    samples=[os.path.basename(file).replace(extension,"") for file in files]
    
    # remove the pair_idenifier from the sample name, if provided
    if pair_identifier:
        # only remove the last instance of the pair identifier
        samples=[pair_identifier.join(sample.split(pair_identifier)[:-1]) if pair_identifier in sample else sample for sample in samples]
    
    if convert:
        samples=samples[0]
    
    return samples

def find_files(folder, extension=None, exit_if_not_found=None):
    """ Return the files in the given folder with the extension if provided
    
    Args:
        folder (string): A path to a folder
        extension (string): The file extension to search for (optional)
        exit_if_not_found (bool): Indicator to check if files exist (optional) 
        
    Requires:
        None
        
    Returns:
        list: A list of files in the folder

    Example:
        files = find_files("examples","fastq")
    """
    
    # get all of the files in the folder
    files=[os.path.join(folder,file) for file in os.listdir(folder)]
    files=list(filter(lambda file: os.path.isfile(file),files))
    
    # filter to only files with extension
    if extension:
        files=list(filter(lambda file: file.endswith(extension), files))
    
    if exit_if_not_found:
        if not files:
            message="ERROR: No files were found in the folder " + folder
            if extension:
                message+=" with extension "+extension
            sys.exit(message+" .\n")
            
    return files

def name_files(names, folder, subfolder=None, tag=None, extension=None, create_folder=None):
    """ Return a list of file names based on the names and folders provided
    
    Args:
        names (list or string): A list of basenames or files.
        folder (string): The path to the folder.
        subfolder (string): The subfolder to use with the files (optional).
        tag (string): The tag to add to the file basenames (optional).
        extension (string): The extension to use for the files (optional).
        create_folder (bool): Create the folder and subfolder if they do not exist (optional).

    Requires:
        None
        
    Returns:
        list: A list of file names (or string if input is string).
        
    Example:
        files = name_files(["file1","file2"], "output")
    """
    
    # if names is a list, convert to string
    was_string=False
    if isinstance(names, str):
        was_string=True
        names=[names]
    
    # get the basenames from the files
    names=[os.path.basename(name) for name in names]
    
    # use the full path to the folder
    folder=os.path.abspath(folder)
    
    # get the name of the full folder plus subfolder if provided
    if subfolder:
        folder=os.path.join(folder,subfolder)

    # add the extension if provided, and replace existing
    if extension:
        names=[os.path.splitext(name)[0]+"."+extension for name in names]

    # add the tag to the names, if provided
    if tag:
        names=[os.path.splitext(name)[0]+"_"+tag+os.path.splitext(name)[1] for name in names]
        
    files=[os.path.join(folder,name) for name in names]
    
    if create_folder:
        create_folders(os.path.dirname(files[0]))
        
    # if the input was originally a string, convert from list
    if was_string:
        files=files[0]
        
    return files

def create_folders(folder):
    """ Create folder if it does not exist
    
    Args:
        folder (string): The full path to the folder.
        
    Requires:
        None
        
    Returns:
        None
        
    Example:
        create_folders("new_folder")
    """
    
    try:
        if not os.path.exists(folder):
            os.makedirs(folder)
    except EnvironmentError:
        print("Warning: Unable to create folder: "+ folder)
        
def match_files(files1,files2,mapping):
    """ Match files from two sets using the mapping provided
    
    Args:
        files1 (list): A list of files for set1
        files2 (list): A list of files for set2
        mapping (string): The file with the mapping information. This file
            should be tab delimited. It can have headers starting with "#". 
            It should have the basenames for the files for set1 and set2 with
            each line as "fileA\tfileB" with fileA in set files1 and fileB in
            set files2.
            
    Requires:
        None
    
    Returns:
        (list): An ordered list of the first set of files
        (list): An ordered list of the second set of files
        The two lists will be the same length with pairs having the same index
            in each list.
        
    Example:
        match_files(["wts_1.fastq","wts_2.fastq"],["wms_1.tsv","wms_2.tsv"],"mapping.tsv")
        
        mapping.tsv contains:
        # wts   wms
        wts_1   wms_1
        wts_2   wms_2

    """
    
    # read in the mapping file
    set_mappings={}
    try:
        file_handle=open(mapping,"r")
        lines=file_handle.readlines()
        file_handle.close()                
    except EnvironmentError:
        sys.exit("ERROR: Unable to read mapping file: " + mapping)
    
    for line in lines:
        if not line.startswith("#"):
            data=line.rstrip().split("\t")
            if len(data) > 1:
                item1=data[0]
                item2=data[1]
                if "." in item1 or "." in item2:
                    sys.exit("ERROR: Sample names should not contain file extensions ('.fastq'),"+
                    "pair identifiers ('.R1.'), or periods ('.') as part of the name.")
                # check for duplicate mappings
                if item1 in set_mappings:
                    print("Warning: Duplicate mapping in file: " + item1)
                set_mappings[item1]=item2
    
    pair1=[]
    pair2=[]
    for item1,item2 in set_mappings.items():
        file1=list(filter(lambda file: os.path.basename(file).startswith(item1),files1))
        file2=list(filter(lambda file: os.path.basename(file).startswith(item2), files2))
        if len(file1) == 1 and len(file2) == 1:
            # check for the pair
            pair1.append(file1[0])
            pair2.append(file2[0])
        elif len(file1) == 0:
            print("Warning: Unable to find file with key, " + item1 + " in folder " + os.path.dirname(files1))
        elif len(file2) == 0:
            print("Warning: Unable to find file with key, " + item2 + " in folder " + os.path.dirname(files2))
        else:
            print("Warning: Duplicate files found for mapping keys: " + item1 + " " + item2)

    if len(pair1) != len(files1):
        print("Warning: Unable to find matches for all of the files in the set.")
    
    return pair1, pair2

def row_average(data):
    """ Compute the average of each row in a data set
        
        Args:
            data (list of lists): Each list in data represents a row of data. 
                
        Requires:
            None
        
        Returns:
            (list): A list of averages, one for each row in the original data.
            
        Example:
            row_average([[1,2,3],[4,5,6]])
    """    
    
    return [sum(row)/(len(row)*1.0) for row in data]

def row_variance(data):
    """ Compute the variance of each row in a data set
        
        Args:
            data (list of lists): Each list in data represents a row of data. 
                
        Requires:
            None
        
        Returns:
            (list): A list of variances, one for each row in the original data.
            
        Example:
            row_variance([[1,2,3],[4,5,6]])
    """
      
    data_averages=row_average(data)
    data_variances=[]
    for average, row in zip(data_averages,data):
        data_variances.append(sum((i-average)**2 for i in row)/(len(row)*1.0))
        
    return data_variances

def relative_abundance(data, percent=False):
    """ Compute the relative abundance values for a set of data 
        
        Args:
            data (list of lists): Each list in data represents a row of data. 
            percent (bool): Abundance is a percent of 100 (30 for 30% instead of 0.3)
                
        Requires:
            None
        
        Returns:
            (list of lists): Each list in data represents a row of data with relative abundance values.
            
        Example:
            relative_abundance([[1,2,3],[4,5,6]])   
    """ 

    # compute the sum for each column
    sums=[0.0]*len(data[0])
    for i in range(len(data[0])):
        for row in data:
            sums[i]+=float(row[i])
            
    relab=[]
    for row in data:
        new_row=[]
        for i, value in enumerate(row):
            try:
                new_value=value/sums[i]
            except ZeroDivisionError:
                new_value=0
            new_row.append(new_value)
        if percent:
            new_row = map(lambda x: x * 100.0, new_row)
        relab.append(new_row)
        
    return relab

def filter_zero_rows(taxa, data, ignore_index=None):
    """ Remove any taxa and data rows from the lists if the data sum for a row is zero.
        
        Args:
            taxa (list): The list of taxa.
            data (list of lists): Each list in data represents a row of data.
            ignore_index (int): An index to ignore in each row in computing the sum.
                
        Requires:
            None
        
        Returns:
            (list): A list of labels for the non-zero rows.
            (list of lists): Each list in data represents a row of data that is non-zero.  
    """ 
    new_taxa=[]
    new_data=[]
    for taxon, row in zip(taxa, data):
        if ignore_index is not None:
            temp_row = row
            del temp_row[ignore_index]
            row_sum = sum(temp_row)
        else:
            row_sum = sum(row)
        if row_sum != 0:
            new_taxa.append(taxon)
            new_data.append(row)
            
    return new_taxa, new_data

def taxa_shorten_name(taxa, level, remove_identifier=None):
    """ Shorten the taxa name by removing the levels indicated (useful for plotting)
        
        Args:
            taxa (list): The list of taxa.
            level (int): The level to filter.
            remove_identifier (bool): If set remove the [k|p|c|r|f|g|s|t__]) from the name. 
                
        Requires:
            None
        
        Returns:
            (list): The list of taxa after removing the unclassified names.  
    """ 

    new_names=[]
    for taxon in taxa:
        name=taxon.split(";")[level]
        if remove_identifier:
            name=name.split("__")[-1]
        new_names.append(name)
        
    return new_names

def top_rows(row_labels, data, max_sets, function):
    """ Get the top rows in the data based on the metric provided 
        
        Args:
            row_labels (list): A list of labels for each row.
            data (list of lists): Each list in data represents a row of data. 
            max_sets (int): Total number of top rows to return.
            function (string): The function to run to get the top values (average or variance)
                
        Requires:
            None
        
        Returns:
            (list): A list of labels for the top rows.
            (list of lists): Each list in data represents a row of data for the top data.
            
        Example:
            top_rows(["row1","row2"],[[1,2,3],[4,5,6]],1)
    """
    
    # get the data after applying the metric function
    if function == "variance":
        stats_data=row_variance(data)
    else:
        stats_data=row_average(data)
    
    # sort the numbers by decreasing order
    sorted_indexes=sorted(range(len(stats_data)),key=lambda i: stats_data[i],reverse=True)
    
    # reduce max sets if the number of rows is less than max sets
    if len(row_labels) < max_sets:
        max_sets = len(row_labels)
    
    top_labels=[]
    top_data=[]
    for i in range(max_sets):
        top_labels.append(row_labels[sorted_indexes[i]])
        top_data.append(data[sorted_indexes[i]])
        
    return top_labels, top_data

def remove_stratified_pathways(pathways, data, remove_description=None):
    """ Remove the stratified pathways from the data set.
        Also remove the unintegrated and unmapped values. Remove the descriptions
        from the pathway names if set.
    
        Args:
            pathways (list): A list of pathway names for each row.
            data (list of lists): Each list in data represents a row of data. 
            remove_description (bool): If set, remove the pathway description
                from the names returned.
                
        Requires:
            None
        
        Returns:
            (list): A list of pathway names.
            (list): A list of lists of the data.
            
        Example:
            remove_stratified_pathways(["pwy1","pwy1|bug1"],[[1,2,3],[4,5,6]])
    """
    
    new_pathways=[]
    new_data=[]
    
    for path, row in zip(pathways, data):
        if not "|" in path and not "UNINTEGRATED" in path and not "UNMAPPED" in path:
            if remove_description:
                path=path.split(":")[0]
            new_pathways.append(path)
            new_data.append(row)
            
    return new_pathways, new_data 

def pathway_names(pathways):
    """ Split the pathway names and descriptions
    
        Args:
            pathways (list): A list of pathway names and descriptions 
                (from a pathway abundance or coverage file)  
        
        Requires:
            None
            
        Returns:
            (dict): A dictionary of pathway names to descriptions
            
    """
    
    path_names = {}
    for path in pathways:
        # ignore stratified pathways
        if not "|" in path:
            try:
                description = path.split(":")
                name = description.pop(0)
                description=":".join(description)
            except ValueError:
                continue
            
            path_names[name]=description
        
    return path_names
        

def filter_taxa_abundance(taxonomy, data, min_abundance, min_samples):
    """ Remove the taxons by min abundance and min samples.
    
        Args:
            taxonomy (list): A list of taxonomy strings for each row.
            data (list of lists): Each list in data represents a row of data. 
            min_abundance (float): Remove data without min abundance. 
            min_samples (float): Remove data not in min samples.
                
        Requires:
            None
        
        Returns:
            (list): A list of species names.
            (list): A list of lists of the data.
            
        Example:
            filter_taxa_abundance(["g__ABC","s__DEF"],[[1,2,3],[4,5,6]],10,2)
    """ 

    filtered_data=[]
    filtered_taxonomy=[]
    # compute the min samples required for this data set
    min_samples_required=math.ceil(len(data[0])*(min_samples/100.0))
    for taxon, data_row in zip(taxonomy, data):
        # filter the species to only include those with min abundance in min of samples
        total_samples_pass_filter=len(list(filter(lambda x: x>min_abundance, data_row)))
        if total_samples_pass_filter >= min_samples_required: 
            filtered_taxonomy.append(taxon)
            filtered_data.append(data_row)
    
    return filtered_taxonomy, filtered_data

def filter_taxa_level_metaphlan2_format(taxonomy, data, min_abundance=None, min_samples=None, level=6):
    """ Remove the taxons that are not a species level (or set a different level with keyword) from the data set.
        Also filter the species if filters are provided. Metaphlan2 format with "|" delimiters and tiered
        abundances (so genus level is split and repeated stratified by species).
    
        Args:
            taxonomy (list): A list of taxonomy strings for each row.
            data (list of lists): Each list in data represents a row of data. 
            min_abundance (float): If set, remove data without min abundance. To
                be used with min_samples.
            min_samples (float): If set, remove data not in min samples.
            level (int): Taxonomic level (default set to species)
                
        Requires:
            None
        
        Returns:
            (list): A list of species names.
            (list): A list of lists of the data.
            
        Example:
            filter_taxa_level_metaphlan2_format(["g__ABC","s__DEF"],[[1,2,3],[4,5,6]])
    """

    taxonomic_levels=["|k__","|p__","|c__","|o__","|f__","|g__","|s__","|t__"]

    # get the level indicator to search for and the next level to filter
    search_taxa = taxonomic_levels[level]
    remove_taxa = taxonomic_levels[level+1] if level+1 < len(taxonomic_levels) else "\n"

    species_data=[]
    species_taxonomy=[]
    # identify the species data in the data set
    # filter out those with species and strain information
    for taxon, data_row in zip(taxonomy, data):
        if search_taxa in taxon and not remove_taxa in taxon:
            species_taxonomy.append(taxon.split("|")[-1].replace(search_taxa[1:],"").replace("_"," "))
            species_data.append(data_row)

    # if filters are provided, then filter the data by both min abundance
    # and min samples
    if min_abundance is not None and min_samples is not None:
        species_taxonomy, species_data = filter_taxa_abundance(species_taxonomy, species_data, min_abundance, min_samples)

    return species_taxonomy, species_data

def read_otu_table(file):
    """ Read in an otu table. Remove extra brackets from taxonomy names if present.
    
        Args:
            file (string): A file containing the otu table (tsv format).
                
        Requires:
            None
        
        Returns:
            (list): A list of samples.
            (list): A list of otu ids.
            (list): A list of taxons.
            (list): A list of lists of data.
            
        Example:
            samples, ids, taxonomy, data = read_otu_table("otu_table.tsv")
    """
        
    data=[]
    samples=[]
    taxonomy=[]
    ids=[]
    with open(file) as file_handle:
        samples = file_handle.readline().rstrip().split("\t")[1:-1]
        for line in file_handle:
            data_points=line.rstrip().split("\t")
            ids.append(data_points.pop(0))
            taxonomy.append(data_points.pop().replace("[","").replace("]",""))
            data.append([float(i) for i in data_points])
            
    return samples, ids, taxonomy, data
    

def sort_data(data, samples, sort_by_name=False, sort_by_name_inverse=False):
    """ Sort the data with those with the largest values first or by sample name

        Args:
            data (list): The data points for each sample.
            samples (list): The sample names that correspond to each data point. 
            sort_by_name (bool): If true, sort by sample name
            sort_by_name_inverse (bool): If true, sort by the inverse of the name (so the reverse of the string)
                this is useful for samples with sample name plus features

        Requires:
            None
        
        Returns:
            (list): The data points for each sample sorted.
            (list): The sample names that correspond to each data point sorted.
            
    """
    import numpy
    
    # if the data is a list of lists of single values, then convert to a list of values
    if isinstance(data[0], list) and max([len(row) for row in data]) == 1:
        data=[row[0] for row in data]
       
    # if set, sort by sample name (samples as columns in the data)
    if sort_by_name or sort_by_name_inverse:
        data_by_sample={sample:data_point for sample,data_point in zip(samples,numpy.transpose(data))}
        sorted_samples=sorted(samples)
        if sort_by_name_inverse:
            # use the reverse of the sample names to sort
            sorted_samples=sorted(samples, key=lambda x: x[::-1])
        sorted_data_transpose=[data_by_sample[sample] for sample in sorted_samples]
        sorted_data = numpy.transpose(sorted_data_transpose)
    else:
        data_by_sample={sample:data_point for sample,data_point in zip(samples,data)}
        sorted_samples=sorted(data_by_sample,key=data_by_sample.get, reverse=True)
        sorted_data=[data_by_sample[sample] for sample in sorted_samples]
   
    return sorted_samples, sorted_data

def is_paired_table(file):
    """ Check if a file contains paired read counts using the header information.
    
        Args:
            file (string): A file of read counts.
                
        Requires:
            None
        
        Returns:
            (bool): True if the file contains paired read counts
            
    """
    
    # read in the first line in the file
    try:
        with open(file) as file_handle:
            header=file_handle.readline()
    except EnvironmentError:
        sys.exit("Unable to read file: " + file)
        
    paired = True if "pair" in header.lower() else False
    
    return paired

def microbial_read_proportion_multiple_databases(data, columns, orphan_data=None, rna=None):
    """ Compute microbial read proportions from the KneadData read counts for 
        multiple databases 
        
        Args:
            data (list of lists): The single or paired data read counts for each sample.
            columns (list): The names of the columns corresponding to the paired data. These
                columns include the reference database names.
            orphan_data (list of lists): The orphan data (if paired end reads)
            rna (bool): If set, this data set is RNA so compute additional ratio.
        
        Requires:
            None
        
        Returns:
            (list of lists): A list of ratios for each sample.
            (list): A list of strings with labels for the ratios.
    """
    
    # compute ratios for each database used for qc
    dna_microbial_reads=[]
    dna_microbial_labels=[]
    for index, qc_database in enumerate(columns[2:]):
        # get a subset of the data for this ratio
        data_subset=[row[:2]+[row[index+2]] for row in data]

        # create subset of orphan data if provided
        orphan_subset=None
        if orphan_data and len(orphan_data[0]) > 2:
            orphan_subset=[row[:2]+[row[index+2]] for row in orphan_data]
        else:
            orphan_subset=orphan_data
        
        reads_ratio, ratio_labels = microbial_read_proportion(data_subset, 
            orphan_data=orphan_subset, database_name=qc_database, rna=rna)
        dna_microbial_labels+=ratio_labels
        if not dna_microbial_reads:
            dna_microbial_reads=reads_ratio
        else:
            dna_microbial_reads=[row1+row2 for row1, row2 in zip(dna_microbial_reads,reads_ratio)]
            
    return dna_microbial_reads, dna_microbial_labels

def microbial_read_proportion(paired_data, orphan_data=None, rna=None, database_name=None):
    """ Compute microbial read proporations from the KneadData read counts.
    
        Args:
            paired_data (list of lists): The paired data read counts for each sample.
            orphan_data (list of lists): The orphan data read counts for each sample. 
            rna (bool): If set, this data set is RNA so compute additional ratio.
            database_name (string): The name of the contaminate database.
                
        Requires:
            None
        
        Returns:
            (list of lists): A list of ratios for each sample.
            (list): A list of strings with labels for the ratios.
            
    """
    
    # if the database name is not set, use the default
    if database_name is None:
        database_name="hg38"
    
    # if the orphan reads are not provided, create an empty set of data
    if orphan_data is None:
        orphan_data=[]
        for i in range(len(paired_data)):
            orphan_data.append([0,0,0,0])
    
    proportion_decontaminated = []
    for paired_row, orphan_row in zip(paired_data, orphan_data):
        decontaminated_sum = 2.0 * paired_row[-1] + orphan_row[-1] + orphan_row[-2]
        decon_trim = decontaminated_sum / (2.0 * paired_row[1] + orphan_row[0] + orphan_row[1])
        decon_raw = decontaminated_sum / (2.0 * paired_row[0])
        if rna:
            decon_ratio = decontaminated_sum / (2.0 * paired_row[-2] + orphan_row[-3] + orphan_row[-4])
            proportion_decontaminated.append(["{0:.5f}".format(i) for i in [decon_trim, decon_ratio, decon_raw]])
        else:
            proportion_decontaminated.append(["{0:.5f}".format(i) for i in [decon_trim, decon_raw]])

    if rna:
        labels=[database_name+" mRNA / Trim",database_name+" mRNA / "+database_name,database_name+" mRNA / Raw"]
    else:
        labels=[database_name+" / Trim",database_name+" / Raw"]
        
    return proportion_decontaminated, labels


def taxa_remove_unclassified(taxa, delimiter=";"):
    """ Rename the taxa to remove the unclassified levels
    
        Args:
            taxa (list): The list of taxa.
            delimiter (str): The string delimiter (usually pipe or semi-colon).
                
        Requires:
            None
        
        Returns:
            (list): The list of taxa after removing the unclassified names.
    """
    
    # remove any levels where the name is unknown (ie empty)
    for taxon in taxa:
        new_name=[]
        for level in taxon.replace(" ","").split(delimiter):
            try:
                rank, name = level.split("__")
            except ValueError:
                # ignore identities like "unclassified" if present
                continue
            if name:
                new_name.append(level)
            else:
                break
        yield delimiter.join(new_name)
        
def taxonomy_trim(taxa):
    """ Trim the taxonomy name to include the most specific known name followed by unclassified
    
        Args:
            taxa (list): The list of taxa.
                
        Requires:
            None
        
        Returns:
            (list): The list of taxa after trimming.
    """
    
    # remove any spaces from the taxonomy
    taxa = [taxon.replace(" ","") for taxon in taxa]

    # determine the delimiter (pipe or semi-colon)
    delimiter="|" if "|" in taxa[0] else ";"
    
    # get the taxa with unclassified levels removed
    taxa_unclassified_removed = taxa_remove_unclassified(taxa,delimiter)
    
    trimmed_taxa=[]
    for taxon_full, taxon_reduced in zip(taxa, taxa_unclassified_removed):
        # if the taxon is specific to species level, then 
        # return the genus and species level
        if taxon_full == taxon_reduced:
            data = taxon_full.split(delimiter)
            trimmed_taxa.append(data[-2]+"."+data[-1])
            
        else:
            most_specific_clade = taxon_reduced.split(delimiter)[-1]
            if not most_specific_clade:
                trimmed_taxa.append(taxon_reduced.replace(delimiter,".")) 
            else:
                data = taxon_full.split(most_specific_clade)
                trimmed_taxa.append(most_specific_clade+data[-1].replace(delimiter,"."))
            
    return trimmed_taxa
        
def terminal_taxa(taxa, data):
    """ Reduce the list of taxa to just those that represent the terminal nodes. If there
        are duplicate terminal nodes, then sum the duplicates.
    
        Args:
            taxa (list): The list of taxa (in the same order as the data).
            data (list of lists): The data points for all samples for each taxa.
                
        Requires:
            None
        
        Returns:
            (list): The list of taxa (terminal node only)
            (list of lists): The data after reducing to terminal node taxa.
    """    
    
    terminal_node_taxa=[]
    # check the taxa by level, starting with the most specific level of strain
    # use a full match with strain instead of just startswith to allow for unclassified
    # strains to not match with classified strains
    # if strains are not present, then run at a species level instead
    
    # check for the most specific taxonomy level (ie strain or species)
    max_taxonomy_level=max([len(taxon.split(";")) for taxon in taxa])
    
    taxa_for_level, data_level=taxa_by_level(taxa, data, level=max_taxonomy_level-1, keep_unclassified=True)
    for taxon in taxa_for_level:
        matching_taxa=list(filter(lambda x: x.replace(" ","") == taxon.replace(" ",""), terminal_node_taxa))
        if len(matching_taxa) == 0:
            terminal_node_taxa.append(taxon)
    
    for level in reversed(range(max_taxonomy_level-1)):
        taxa_for_level, data_level=taxa_by_level(taxa, data, level, keep_unclassified=True)
        for taxon in taxa_for_level:
            # check if part of this taxon is already included
            matching_taxa=list(filter(lambda x: x.replace(" ","").startswith(taxon.replace(" ","")), terminal_node_taxa))
            if len(matching_taxa) == 0:
                terminal_node_taxa.append(taxon)
                
    # create a set of terminal node taxa and data
    new_taxa={}
    for taxon, row in zip(taxa, data):
        if taxon in terminal_node_taxa:
            if taxon in new_taxa:
                new_taxa[taxon]=[a+b for a,b in zip(new_taxa[taxon],row)]
            else:
                new_taxa[taxon]=row
               
    new_taxa_list=sorted(new_taxa.keys())
    new_data_list=[new_taxa[i] for i in new_taxa_list] 
               
    return new_taxa_list, new_data_list
                    
def taxa_by_level(taxa, data, level, keep_unclassified=None):
    """ Combine the data to represent the taxa by a specific level
    
        Args:
            taxa (list): The list of taxa (in the same order as the data).
            data (list of lists): The data points for all samples for each taxa.
            level (int): The level to sum the taxa (zero is kingdom level).
            keep_unclassified (bool): If set, keep unclassified taxa.
                
        Requires:
            None
        
        Returns:
            (list): The list of taxa (all to the level specified)
            (list of lists): The data after summing to the taxa level specified.
    """    

    # first remove any unclassified levels
    if not keep_unclassified:
        taxa=taxa_remove_unclassified(taxa)
    
    # sum the taxa by the level provided
    data_sum={}
    for taxon, taxon_data in zip(taxa, data):
        split_taxon=taxon.split(";")
        if len(split_taxon) < (level+1):
            # do not include those taxa that are not specified to the level requested
            continue
        new_taxon_level=";".join(split_taxon[:(level+1)])
        if new_taxon_level in data_sum:
            data_sum[new_taxon_level]=[a+b for a,b in zip(data_sum[new_taxon_level],taxon_data)]
        else:
            data_sum[new_taxon_level]=taxon_data
        
    new_taxa=[]
    new_data=[]    
    for taxon, taxon_data in data_sum.items():
        new_taxa.append(taxon)
        new_data.append(taxon_data)
        
    return new_taxa, new_data

def format_data_comma(data):
    """ Format the numbers in the string to include commas.
    
    Args:
        data (string or list): A text string.
        
    Requires:
        None
        
    Returns:
        (string): A text string.
        
    """
    
    if not isinstance(data,list):
        data=data.split()

    new_string=[]
    for token in data:
        try:
            new_token="{:,}".format(int(token))
        except ValueError:
            new_token=token
        new_string.append(new_token)
        
    return " ".join(new_string)

def read_eestats2(file):
    """ Read the eestats2 file which is an ascii table.
    
    Args:
        file (string): The path to the eestats file.
        
    Requires:
        None
        
    Returns:
        (list): The table rows.
        (list): The table columns.
        (list): The table data.
        (string): The summary.
        
    """
    
    with open(file) as file_handle:
        eestats_lines = file_handle.readlines()
        
    # read in the overall stats line
    overall_stats = format_data_comma(eestats_lines[1].rstrip())
    # read in the maxee values from the columns
    columns = list(filter(lambda x: x.strip() and not x in ["Length","MaxEE"],eestats_lines[3].rstrip().split()))
    columns = [column+" maxee" for column in columns]
    rows = []
    data = []
    # read through the data table
    for line in eestats_lines[5:]:
        stats = list(filter(lambda x: x.strip(),line.strip().split("   ")))
        # move spaces in data values and percents
        stats = [stat.replace("(  ","(").replace("( ","(").replace("("," (") for stat in stats]
        rows.append(stats.pop(0)+" nt")
        data.append([format_data_comma(stat) for stat in stats])
        
    return rows, columns, data, overall_stats

def get_files(folder, extension):
    """ Return paths to all files in a folder with a given extension """
    
    for file in os.listdir(folder):
        file=os.path.join(folder, file)
        if os.path.isfile(file) and file.endswith(extension):
            yield file

def read_picard(file, threshold=20):
    """ Read the picard file which is an ascii table of quality scores per base.
    
    Args:
        file (string): The path to the picard file.
        
    Requires:
        None
        
    Returns:
        (list): A list of tuples of base / quality score. 
        (bool): True if any quality score in the set does not meet the threshold.
    """
    
    with open(file) as file_handle:
        picard_lines = file_handle.readlines()
        
    # read through file to get quality scores ignoring all lines with comments/headers
    data=[]
    below_threshold=False
    for line in picard_lines:
        if not (line.startswith("#") or line.startswith("CYCLE")):
            new_data=line.rstrip().split("\t")
            # check if the quality score passes the threshold
            try:
                new_data=(int(new_data[0]),float(new_data[1]))
                if new_data[1] < threshold:
                    below_threshold=True
                data.append(new_data)
            except ValueError:
                pass
        
    return data, below_threshold

def rank_species_average_abundance(file):
    """ Read in a taxonomy file, and sort species by average abundance 
    
    Args:
        file (string): The path to the merged taxonomy file generated by MetaPhlAn2
        
    Requires:
        None
        
    Returns:
        (list): A list of species ordered by decreasing average abundance
    """
    
    def try_format_data(value):
        """ Try to format the data in a file """
        try:
            value = float(value)
        except ValueError:
            value= 0
            
        return value
    
    species=collections.OrderedDict()
    with open(file) as file_handle:
        try:
            column_names = file_handle.readline().rstrip().split("\t")[1:]
        except IndexError:
            column_names = []
        for line in file_handle:
            line=line.rstrip().split("\t")
            taxonomy=line.pop(0).split("|")[-1]
            data=[try_format_data(i) for i in line]
            try:
                average=sum(data)/(len(data)*1.0)
            except ZeroDivisionError:
                average=0
            # only store values for species
            if taxonomy.startswith("s__"):
                species[taxonomy]=average
                
    # sort the species from highest to lowest average abundance
    sorted_species = sorted(species, key=species.get, reverse=True)
    # if abundances are not provided then use original ordering
    if sum(species.values()) == 0:
        sorted_species = species.keys()
    
    return sorted_species

def order_clade_list(task,clade_list,abundance_file,output_file):
    """ Using the average abundances order the clade list from strainphlan 
    
    Args:
        task (anadama2.task): An instance of the task class.
        clade_list (string): The path to the file containing the strainphlan list of clades.
        abundance_file (string): The path to the merged abundance file from metaphlan2.
        output_file (string): The file to write the ordered clade list.
        
    Requires:
        none
        
    Returns:
        none
    """
    
    # get the species listed by average abundance
    species_ranked = rank_species_average_abundance(abundance_file)
    
    # read in the clade list
    clades=set()
    with open(clade_list) as file_handle:
        for line in file_handle:
            if line.startswith("s__"):
                clades.add(line.rstrip().split(" ")[0])
        
    # write out ordered species also included in clade list
    with open(output_file,"w") as file_handle:
        for taxon in species_ranked:
            if taxon in clades:
                file_handle.write(taxon+"\n")
    

def sort_fastq_file(task):
    """Sorts a FASTQ file by name (sequence identifier contents).

    Args:
        task (anadama2.task): An instance of the task class.

    Requires:
        None

    Returns:
        None
    """
    sample_name = os.path.basename(task.depends[0].name)
    output_dir = os.path.dirname(task.targets[0].name)
    temp_dir = os.path.join(output_dir, "%s.tmp" % sample_name)

    if task.depends[0].name.endswith('.gz'):
        sort_command = "zcat "
    else:
        sort_command = "cat "        
    
    sort_command += ("[depends[0]] | paste - - - - | sort -T [depends[1]] -k1,1 | "
                    "tr '\t' '\n' > [targets[0]]")

    run_task('mkdir -p [targets[0]]',
             depends=[TrackedDirectory(output_dir)],
             targets=[TrackedDirectory(temp_dir)])

    run_task(sort_command,
             depends=task.depends + [TrackedDirectory(temp_dir), TrackedDirectory(output_dir)],
             targets=task.targets)

    run_task("rm -rf [depends[0]]",
             depends=[TrackedDirectory(temp_dir)] + task.targets)


def extract_orphan_reads(task):
    """Extracts orphan reads from the provided input files. Orphan reads are saved into a separate file 
    for further downstream analysis.

    Args:
        task (anadama2.task): An instance of the task class.

    Requires:
        seqtk v1.2+: A fast and lightweight tool for processing sequences in the FASTA
             or FASTQ format

    Returns:
        None
    """
    sample_name = os.path.basename(os.path.dirname(task.depends[0].name))
    orphans_dir = os.path.dirname(task.targets[0].name)

    run_task("seqtk dropse [depends[0]] > [targets[0]]",
             depends=task.depends,
             targets=task.targets[0])

    run_task("extract_orphan_reads.py -r [depends[0]] -b [depends[1]] -o [depends[2]]",
             depends=[task.depends[0], task.targets[0], TrackedDirectory(orphans_dir)],
             targets=[task.targets[1]],
             args=[".fastq"])


def is_paired_end(input_files, extension, pair_identifier):
    """Returns true if the provided set of input files are paired-end.

    Args:
        input_files (list): A list of files to verify if paired-end data.
        extension (string): The extension for all files.
        pair_identifier (string): The string in the file basename to identify
            the first pair in the set.

    Requires:
        None

    Returns:
        bool: True if paired-end datasets; False otherwise.
    """
    input_pair1, input_pair2 = paired_files(input_files, extension, pair_identifier)
    
    return True if input_pair1 else False



#!/usr/bin/env python

# Given a raw interleaved fastq file and a balanced interleaved fastq (containing no orpahns)
# generate the read ID lists necessary to extract all orphans from the original file.
#
# Makes use of the seqtk utility

import argparse
import os
import subprocess


def parse_cli_arguments():
    """
    """
    parser = argparse.ArgumentParser('Extracts orphan reads from a interleaved sequence file '
                                     'and produces an orphan sequence file.')
    parser.add_argument('-r', '--raw-sequence', required=True, 
                        help='The raw interleaved sequence file.')
    parser.add_argument('-b', '--balanced-sequence', required=True,
                        help='Balanced sequence file with no orphan sequences.')
    parser.add_argument('-o', '--output-dir', required=True,
                        help='Output directory to write orphan sequence files too.')    

    return parser.parse_args()


def get_ids_from_sequences(sample_name, raw_seq, balanced_seq, out_dir):
    """
    Extracts raw, balanced and orphan sequence IDs from the provided 
    sequence files.
    """
    raw_ids = os.path.join(out_dir, "%s.raw_ids.txt" % sample_name)
    balanced_ids = os.path.join(out_dir, "%s.matched_ids.txt" % sample_name)
    orphan_ids = os.path.join(out_dir, "%s.orphan_ids.txt" % sample_name)

    for (input_seq, output_ids) in [(raw_seq, raw_ids), (balanced_seq, balanced_ids)]:
        with open(output_ids, 'wb') as out_ids:
            ps_grep = subprocess.Popen(['grep', '-e', '^@.*/[1|2]$', input_seq], stdout=subprocess.PIPE)
            ps_sed = subprocess.Popen(['sed', '-e', 's/^@//'], stdin=ps_grep.stdout, stdout=subprocess.PIPE)
            ps_grep.stdout.close()

            ps_sort = subprocess.Popen(['sort'], stdin=ps_sed.stdout, stdout=out_ids)
            ps_sed.stdout.close()

            ps_sort.communicate()

    with open(orphan_ids, 'wb') as orphan_ids_out:
        p = subprocess.Popen(['comm', '-23', raw_ids, balanced_ids], stdout=orphan_ids_out)
        p.communicate()

    return (raw_ids, balanced_ids, orphan_ids)


def generate_orphan_sequences(sample_name, raw_seqs, orphan_ids, out_dir):
    """
    Generates an orphan sequence file from the supplied interleaved sequence file.
    """
    orphan_seqs_file = os.path.join(out_dir, "%s_orphans.fastq" % sample_name)

    with open(orphan_seqs_file, 'wb') as orphan_seqs:
        p = subprocess.Popen(['seqtk', 'subseq', raw_seqs, orphan_ids], stdout=orphan_seqs)
        p.communicate()


def main(args):
    sample_name = os.path.basename(args.raw_sequence).split(os.extsep, 1)[0]

    (raw_ids, balanced_ids, orphan_ids) = get_ids_from_sequences(sample_name,
                                                                 args.raw_sequence, 
                                                                 args.balanced_sequence,
                                                                 args.output_dir)

    generate_orphan_sequences(sample_name, args.raw_sequence, orphan_ids, args.output_dir)

    os.remove(raw_ids)
    os.remove(balanced_ids)
    os.remove(orphan_ids)


if __name__ == "__main__":
    main(parse_cli_arguments())




import os

def get_file(file):
    # get the full path to the data file
    install_folder=os.path.dirname(os.path.realpath(__file__))
    return os.path.join(install_folder,file)

def get_tutorial_folder():
    return os.path.join(os.path.dirname(os.path.realpath(__file__)),
        os.pardir,os.pardir,"tutorial")

def get_kneaddata_hg_demo_folder():
    return os.path.join(get_tutorial_folder(),"kneaddata_demo_db","Homo_sapiens_demo")

def get_kneaddata_silva_demo_folder():
    return os.path.join(get_tutorial_folder(),"kneaddata_demo_db","SILVA_demo")


"""
bioBakery Workflows: config module
Configuration settings for workflows and tasks

Copyright (c) 2016 Harvard School of Public Health

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import os
import sys

def get_home_directory():
    """ Try to get the users home directory """
    
    try:
        directory=os.path.expanduser("~")
    except EnvironmentError:
        directory=None
        
    return directory
    
def install_locations():
    """ Get the possible default install locations """
    
    main_folder=Workflow.base_environment_variable.lower()
    options=[os.path.join(os.path.sep,"opt",main_folder)]
    home_dir=get_home_directory()
    if home_dir:
        options+=[os.path.join(home_dir,main_folder)]
        
    return options

class DBInfo(object):
    def __init__(self, name, description, url=None, file_name=None, default_path=None):
        self.name = name
        self.description = description
        self.default_folder = name.lower()
        self.url = url
        
        # if the file name is not set, get it from the url
        if not file_name and url:
            file_name = os.path.basename(url)
        self.file_name = file_name
        
        # set default path if not provided to default folder/file names
        if file_name and not default_path:
            default_path = os.path.join(self.default_folder, file_name)
        elif not default_path:
            default_path = self.default_folder
            
        self.default_path = default_path
        
def get_environment_variable(name):
    """ Try to get the environment variable """

    variable = None
    try:
        variable = os.environ[name]
    except KeyError:
        pass
    
    return variable
    

class Workflow(object):
    base_environment_variable = "BIOBAKERY_WORKFLOWS_DATABASES"
    
    def __getattr__(self, name):
        """ Try to get the database location """
        
        # first check for the environment variable for the database
        variable = get_environment_variable(self.vars[name].name)
        
        if not variable:
            # next check for the overall location variable
            base_folder = get_environment_variable(self.base_environment_variable)
            if base_folder:
                variable = os.path.join(base_folder,self.vars[name].default_path)
                
        if variable and not (os.path.isfile(variable) or os.path.isdir(variable)):
            # if this is not a valid folder/file, then resume search
            variable = None
        
        # check in the default folders, if not found with env variables
        if not variable:
            for folder in install_locations():
                if os.path.isdir(folder):
                    variable = os.path.join(folder,self.vars[name].default_path)
                    if os.path.isfile(variable) or os.path.isdir(variable):
                        return variable
        
        if not variable:
            sys.exit("ERROR: Unable to find database "+self.vars[name].name+
                ". "+self.vars[name].description+" Unable to find in default"+
                " install folders or with environment variables.")
            
        return variable

class ShotGun(Workflow):
    vars={}
    vars["kneaddata_db_human_genome"]=DBInfo("KNEADDATA_DB_HUMAN_GENOME",
        description="This is the KneadData bowtie2 database of the human genome. "+\
            "This database can be downloaded with KneadData.")
    vars["kneaddata_db_human_metatranscriptome"] = DBInfo("KNEADDATA_DB_HUMAN_TRANSCRIPTOME",
        description="This is the folder containing the KneadData bowtie2 database for the human transcriptome."+\
            "This database can be downloaded with KneadData.")
    vars["kneaddata_db_rrna"] = DBInfo("KNEADDATA_DB_RIBOSOMAL_RNA",
        description="This is the folder containing the KneadData bowtie2 database of SILVA rRNA."+\
            "This database can be downloaded with KneadData.")
    vars["strainphlan_db_reference"] = DBInfo("STRAINPHLAN_DB_REFERENCE",
        description="This is the folder containing the reference genomes used "+\
            "when running StrainPhlAn.")
    vars["strainphlan_db_markers"] = DBInfo("STRAINPHLAN_DB_MARKERS",
        description="This is the folder containing the StrainPhlAN marker files.")
    vars["panphlan_db"] = DBInfo("PANPHLAN_DB",
        description="This is the folder containing the PanPhlAn database files.")

class SixteenS(Workflow):
    vars={}
    vars["greengenes_fasta"] = DBInfo("GREEN_GENES_FASTA_DB",
        description="This is the GreenGenes fasta file.",
        url="ftp://greengenes.microbio.me/greengenes_release/gg_13_8_otus/rep_set/97_otus.fasta")
    vars["greengenes_usearch"] = DBInfo("GREEN_GENES_USEARCH_DB",
        description="This is the GreenGenes file formatted for Usearch. "+\
            "A fasta formatted file can be used.",
        file_name=vars["greengenes_fasta"].file_name)
    vars["greengenes_taxonomy"] = DBInfo("GREEN_GENES_TAXONOMY_DB",
        description="This is the GreenGenes taxonomy file matching the fasta file.",
        url="ftp://greengenes.microbio.me/greengenes_release/gg_13_8_otus/taxonomy/97_otu_taxonomy.txt")
    vars["greengenes_dada2"] = DBInfo("GREEN_GENES_DADA2_DB",
        description="This is the GreenGenes dada2 database",
        url="https://zenodo.org/record/158955/files/gg_13_8_train_set_97.fa.gz")
    vars["rdp_dada2"] = DBInfo("RDP_DADA2_DB",
        description="This is the rdp dada2 database",
        url="https://zenodo.org/record/158955/files/rdp_train_set_14.fa.gz")
    vars["silva_dada2"] = DBInfo("SILVA_DADA2_DB",
        description="This is the silva dada2 database",
        url="https://zenodo.org/record/1172783/files/silva_nr_v132_train_set.fa.gz")
    vars["rdp_species_dada2"] = DBInfo("RDP_SPECIES_DADA2_DB",
        description="This is the rdp species assignment dada2 database",
        url="https://zenodo.org/record/801828/files/rdp_species_assignment_16.fa.gz")
    vars["silva_species_dada2"] = DBInfo("SILVA_SPECIES_DADA2_DB",
        description="This is the silva species assignment dada2 database",
        url="https://zenodo.org/record/801832/files/silva_species_assignment_v128.fa.gz")
    vars["unite_zip"] = DBInfo("UNITE", description="This is the UNITE fungi database zip file for ITS workflow",
        url="https://files.plutof.ut.ee/doi/B2/07/B2079372C79891519EF815160D4467BBF4AF1288A23E135E666BABF2C5779767.zip")
    vars["unite"] = DBInfo("UNITE", description="This is the UNITE fungi database fasta for ITS workflow",
                               url="unite/sh_general_release_dynamic_01.12.2017.fasta")




#!/usr/bin/env python

import sys
import os
import argparse

# This script will add the functional annotations from eggnog mapper and run_dbcan to a file of genes.

EGGNOG_KEGG_KO_INDEX = 8
EGGNOG_KEGG_PATHWAY_INDEX = 9
EGGNOG_KEGG_MODULE_INDEX = 10
EGGNOG_CAZY_INDEX = 15
DBCAN_DIAMOND_INDEX = 3

ANNOTATION_DELIMITER = ";"

def parse_arguments(args):
    """ 
    Parse the arguments from the user
    """
    parser = argparse.ArgumentParser(
        description= "Annotate Genome\n",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "-i", "--input-genome",
        help="the fasta file\n[REQUIRED]",
        metavar="<input.fasta>",
        required=True)
    parser.add_argument(
        "-e", "--input-eggnog",
        help="the eggnog annotations\n[REQUIRED]",
        metavar="<input.emapper.annotations>",
        required=True)
    parser.add_argument(
        "-d", "--input-dbcan",
        help="the dbcan annotations\n[REQUIRED]",
        metavar="<input.dbcan.overview.txt>",
        required=True)
    parser.add_argument(
        "-o", "--output",
        help="file to write the annotated genome\n[REQUIRED]",
        metavar="<output.fasta>",
        required=True)

    return parser.parse_args()

def format_cell(item, kegg_ko=False):
    # replace empty cells with none
    # remove ko: from each if kegg cell
    if item == "":
        item = "None"
    if kegg_ko:
        item = item.replace("ko:","")
    return item

def main():
    # read in run parameters
    args=parse_arguments(sys)

    # read in eggnog mapper annotations
    eggnog_kegg={}
    eggnog_cazy={}
    print("Reading eggnog annotations")
    with open(args.input_eggnog) as file_handle:
        for line in file_handle:
            if not line.startswith("#"):
                info = line.rstrip().split("\t")
                eggnog_kegg[info[0]]=(format_cell(info[EGGNOG_KEGG_KO_INDEX],kegg_ko=True),
                    format_cell(info[EGGNOG_KEGG_PATHWAY_INDEX]),
                    format_cell(info[EGGNOG_KEGG_MODULE_INDEX]))
                eggnog_cazy[info[0]]=format_cell(info[EGGNOG_CAZY_INDEX])

    # read in dbcan annotations
    run_dbcan={}
    print("Reading dbcan annotations")
    with open(args.input_dbcan) as file_handle:
        header = file_handle.readline()
        for line in file_handle:
            info = line.rstrip().split("\t")
            run_dbcan[info[0]]=format_cell(info[DBCAN_DIAMOND_INDEX])

    # write new fasta file with annotations
    print("Writing annotated fasta file")
    with open(args.output,"w") as file_handle_write:
        with open(args.input_genome) as file_handle_read:
            for line in file_handle_read:
                if line.startswith(">"):
                    info = line.rstrip().split(" ")
                    gene_id = info[0][1:]
                    name = "_".join(info[1:])
                    kegg_ko, kegg_pathway, kegg_module = eggnog_kegg.get(gene_id,["None","None","None"])
                    kegg = ANNOTATION_DELIMITER.join(["kegg_ko:"+kegg_ko, "kegg_pathway:"+kegg_pathway,"kegg_module:"+kegg_module])
                    cazy = "cazy:"+eggnog_cazy.get(gene_id,"None")
                    dbcan = "dbcan_diamond:"+run_dbcan.get(gene_id,"None")
                    file_handle_write.write(">"+";".join([gene_id,name,kegg,cazy,dbcan])+"\n")
                else:
                    file_handle_write.write(line)
            
if __name__ == "__main__":
    main()






#+ echo=False
max_sets=50

from biobakery_workflows import utilities
from biobakery_workflows import visualizations

from anadama2 import PweaveDocument

document=PweaveDocument()  

# get the variables for this document generation task
vars = document.get_vars()

# determine the document format
pdf_format = True if vars["format"] == "pdf" else False

#' # Functional Profiling of Metagenomic Reads

#' <%= visualizations.ShotGun.captions["functional_intro"] %>

#+ echo=False

# read in the top average pathways
dna_samples, dna_top_average_pathways, dna_top_average_data, top_names_and_descriptions = visualizations.top_average_pathways(
    document, vars["dna_pathabundance"], max_sets)
dna_ecs_samples, dna_top_average_ecs, dna_top_average_ecs_data, top_ecs_names_and_descriptions = visualizations.top_average_pathways(
    document, vars["dna_ecabundance"], max_sets)

#' ## Pathway and ECs Abundance

#' <%= visualizations.ShotGun.format_caption("heatmap_intro",max_sets=max_sets,type="pathways",method="Spearman") %>

#+ echo=False
# update the figure size based on output format for the heatmaps
utilities.change_pweave_figure_size_heatmap(pdf_format)

#+ echo=False
# if there is metadata, add it to the heatmap
def log10_heatmap(dna_samples, dna_top_average_pathways, dna_top_average_data, data_type="pathways"):
    merged_data=[]
    metadata_pathways=[]
    metadata_samples=[]
    if 'metadata' in vars and vars['metadata']:
        merged_data, metadata_samples=utilities.merge_metadata(vars['metadata'], dna_samples, 
            [[dna_top_average_pathways[i]]+dna_top_average_data[i] for i in range(len(dna_top_average_pathways))])
        metadata_pathways=[row.pop(0) for row in merged_data]
        # get the metadata row numbers
        metadata_rows=range(1,len(vars['metadata']))
        document.show_hclust2(metadata_samples, metadata_pathways, merged_data,
            title="Top "+str(max_sets)+" "+data_type+" by average abundance",
            metadata_rows=metadata_rows)
    else:
        document.show_hclust2(dna_samples,dna_top_average_pathways,dna_top_average_data,
            title="Top "+str(max_sets)+" "+data_type+" by average abundance")  

    return merged_data, metadata_pathways, metadata_samples

merged_data, metadata_pathways, metadata_samples=log10_heatmap(dna_samples, dna_top_average_pathways, dna_top_average_data)

#' <%= visualizations.ShotGun.format_caption("pathway_abundance_heatmap",norm="log10") %> 

#+ echo=False
ec_merged_data, ec_pathways, ec_samples=log10_heatmap(dna_ecs_samples, dna_top_average_ecs, dna_top_average_ecs_data, "ecs")

#' <% if pdf_format: print("\clearpage") %>

#+ echo=False
def zscore_heatmap(dna_samples, dna_top_average_pathways, dna_top_average_data, merged_data, metadata_pathways, metadata_samples, data_type="pathways"):
    # if there is metadata, add it to the heatmap
    if 'metadata' in vars and vars['metadata'] and 'metadata_labels' in vars and vars['metadata_labels']:
        # get the total number of features
        total_features=len(vars['metadata'])-1
        # for the zscore to be applied first all of the categorical data must be removed
        filtered_metadata_pathways=[]
        filtered_merged_data=[]
        filtered_row_count=0
        for i in range(total_features):
            label = vars['metadata_labels'].get(metadata_pathways[i],"cat")
            if label != "cat":
                filtered_metadata_pathways.append(metadata_pathways[i])
                filtered_merged_data.append(merged_data[i])
            else:
                filtered_row_count+=1
        filtered_metadata_rows=range(1,total_features-filtered_row_count+1)
    
        # add abundance data to filtered metadata
        filtered_metadata_pathways+=metadata_pathways[total_features:]
        filtered_merged_data+=merged_data[total_features:]
      
        document.show_hclust2(metadata_samples, filtered_metadata_pathways, filtered_merged_data,
            title="Top "+str(max_sets)+" "+data_type+" by average abundance",
            log_scale=False,zscore=True,
            metadata_rows=filtered_metadata_rows)
    else:
        document.show_hclust2(dna_samples,dna_top_average_pathways,dna_top_average_data,
            title="Top "+str(max_sets)+" "+data_type+" by average abundance",
            log_scale=False,zscore=True)

zscore_heatmap(dna_samples, dna_top_average_pathways, dna_top_average_data, merged_data, metadata_pathways, metadata_samples)

#' <%= visualizations.ShotGun.format_caption("pathway_abundance_heatmap",norm="z-score") %> 

#+ echo=False
zscore_heatmap(dna_ecs_samples, dna_top_average_ecs, dna_top_average_ecs_data, ec_merged_data, ec_pathways, ec_samples, "ecs")

#' <% if pdf_format: print("\clearpage") %>

#+ echo=False
# reset the figure size to the defaults
utilities.reset_pweave_figure_size()

#+ echo=False

# write a table of the pathways average and variance
pathway_file_name="top_average_pathways_names.tsv"
average_abundance_variance=visualizations.write_pathway_average_variance_table(document, pathway_file_name, dna_top_average_data, top_names_and_descriptions)

table_message=visualizations.show_table_max_rows(document, average_abundance_variance, 
    top_names_and_descriptions, [" Average "," Variance "], 
    "Top "+str(max_sets)+" pathways by average abundance", pathway_file_name, font=7)

#' <%= table_message %>

#' <% visualizations.print_pathways_urls(dna_top_average_pathways,top_names_and_descriptions,3) %>

#' <% if pdf_format: print("\clearpage") %>

#' ## Features

#' <%= visualizations.ShotGun.format_caption("feature_detection",seq_type="metagenomic",seq_short_type="DNA") %>

#+ echo=False

# read in the read count and feature count files
total_reads, nucleotide_reads, translated_reads, genefamilies_counts, ecs_counts, pathabundance_counts = visualizations.feature_counts(
    document, vars["read_counts"],vars["feature_counts"])

# add scatter plots of the data
document.plot_scatter([[total_reads,nucleotide_reads],[total_reads,translated_reads]],title="Read alignment rate",
                       row_labels=["Nucleotide search","Nucleotide + translated search"],xlabel="log10(Input reads)", ylabel="log10(Aligned reads)", trendline=True)

#' <%= visualizations.ShotGun.captions["scatter_reads_aligned"] %> 

#+ echo=False
document.plot_scatter([[nucleotide_reads,genefamilies_counts],[translated_reads,genefamilies_counts]],title="UniRef90 gene families",
                       row_labels=["Nucleotide search","Nucleotide + translated search"],xlabel="log10(Aligned reads)", ylabel="log10(gene families)", trendline=True)

#+ echo=False
document.plot_scatter([[nucleotide_reads,ecs_counts],[translated_reads,ecs_counts]],title="Enzymes (ECs)",
                       row_labels=["Nucleotide search","Nucleotide + translated search"],xlabel="log10(Aligned reads)", ylabel="log10(ECs)", trendline=True)

#+ echo=False
document.plot_scatter([[nucleotide_reads,pathabundance_counts],[translated_reads,pathabundance_counts]],title="Pathways",
                       row_labels=["Nucleotide search","Nucleotide + translated search"],xlabel="log10(Aligned reads)", ylabel="log10(Pathways)", trendline=True)

#' <%= visualizations.ShotGun.captions["scatter_features"] %>




#+ echo=False
import numpy

from biobakery_workflows import utilities, visualizations, files

from anadama2 import PweaveDocument

document=PweaveDocument()  

# get the variables for this document generation task
vars = document.get_vars()

# determine the document format
pdf_format = True if vars["format"] == "pdf" else False

# read in the DNA samples
(dna_paired_columns, dna_orphan_columns), dna_samples, (dna_paired_data, dna_orphan_data) = visualizations.qc_read_counts(document, vars["dna_read_counts"])

# read in the RNA samples
(rna_paired_columns, rna_orphan_columns), rna_samples, (rna_paired_data, rna_orphan_data) = visualizations.qc_read_counts(document, vars["rna_read_counts"])


#' # Quality Control

#' <% visualizations.ShotGun.print_qc_intro_caption("{} DNA and {} RNA ".format(len(dna_samples),len(rna_samples)), rna_paired_columns[2:], paired=True) %>

#+ echo=False

#' ## DNA Samples Quality Control

#' ### DNA Samples Tables of Filtered Reads

#+ echo=False
document.write_table(["# Sample"]+dna_paired_columns, dna_samples, dna_paired_data,
    files.ShotGunVis.path("qc_counts_paired",document.data_folder))

table_message=visualizations.show_table_max_rows(document, dna_paired_data, dna_samples,
    dna_paired_columns, "DNA Paired end reads", files.ShotGunVis.path("qc_counts_paired"),
    format_data_comma=True)

#' <%= table_message %>

#+ echo=False
document.write_table(["# Sample"]+dna_orphan_columns, dna_samples, dna_orphan_data,
    files.ShotGunVis.path("qc_counts_orphan",document.data_folder))

table_message=visualizations.show_table_max_rows(document, dna_orphan_data, dna_samples,
    dna_orphan_columns, "DNA Orphan reads", files.ShotGunVis.path("qc_counts_orphan"),
    format_data_comma=True)
        
#' <%= table_message %> 

#' <% if pdf_format: print("\clearpage") %>

#+ echo=False
# plot the microbial reads ratios
dna_microbial_reads, dna_microbial_labels = utilities.microbial_read_proportion_multiple_databases(
    dna_paired_data, dna_paired_columns, dna_orphan_data)
document.write_table(["# Sample"]+dna_microbial_labels, dna_samples, 
    dna_microbial_reads, files.ShotGunVis.path("microbial_counts",document.data_folder))

table_message=visualizations.show_table_max_rows(document, dna_microbial_reads, dna_samples,
    dna_microbial_labels, "DNA microbial read proportion",
    files.ShotGunVis.path("microbial_counts"))

#' <%= visualizations.ShotGun.captions["microbial_ratios"] %>  

#' <%= table_message %>

#' ### DNA Samples Plots of Filtered Reads

#+ echo=False
document.plot_grouped_barchart(numpy.transpose(dna_paired_data), row_labels=dna_paired_columns, 
    column_labels=dna_samples, title="DNA Paired end reads", ylabel="Read count (in millions)",
    legend_title="Filter", yaxis_in_millions=True)

#+ echo=False
document.plot_grouped_barchart(numpy.transpose(dna_orphan_data), row_labels=dna_orphan_columns, 
    column_labels=dna_samples, title="DNA Orphan reads", ylabel="Read count (in millions)",
    legend_title="Filter", yaxis_in_millions=True)

#' ## RNA Samples Quality Control

#' ### RNA Samples Tables of Filtered Reads

#+ echo=False
document.write_table(["# Sample"]+rna_paired_columns, rna_samples, rna_paired_data,
    files.ShotGunVis.path("rna_qc_counts_paired",document.data_folder))

table_message=visualizations.show_table_max_rows(document, rna_paired_data, rna_samples,
    rna_paired_columns, "RNA Paired end reads", files.ShotGunVis.path("rna_qc_counts_paired"),
    format_data_comma=True)

#' <%= table_message %>

#+ echo=False
document.write_table(["# Sample"]+rna_orphan_columns, rna_samples, rna_orphan_data,
    files.ShotGunVis.path("rna_qc_counts_orphan",document.data_folder))

table_message=visualizations.show_table_max_rows(document, rna_orphan_data, rna_samples,
    rna_orphan_columns, "RNA Orphan reads", files.ShotGunVis.path("rna_qc_counts_orphan"),
    format_data_comma=True)

#' <%= table_message %>

#' <% if pdf_format: print("\clearpage") %>

#+ echo=False
# write and plot the microbial reads ratios
rna_microbial_reads, rna_microbial_labels = utilities.microbial_read_proportion_multiple_databases(
    rna_paired_data, rna_paired_columns, rna_orphan_data)
document.write_table(["# Sample"]+rna_microbial_labels, rna_samples, 
    rna_microbial_reads, files.ShotGunVis.path("rna_microbial_counts",document.data_folder))

table_message=visualizations.show_table_max_rows(document, rna_microbial_reads, rna_samples,
    rna_microbial_labels, "RNA microbial read proportion",
    files.ShotGunVis.path("rna_microbial_counts"))

#' <%= visualizations.ShotGun.captions["microbial_ratios"] %>  

#' <%= table_message %> 

#' ### RNA Samples Plots of Filtered Reads

#+ echo=False
document.plot_grouped_barchart(numpy.transpose(rna_paired_data), row_labels=rna_paired_columns, 
    column_labels=rna_samples, title="RNA Paired end reads", ylabel="Read count (in millions)",
    legend_title="Filter", yaxis_in_millions=True)

#+ echo=False
document.plot_grouped_barchart(numpy.transpose(rna_orphan_data), row_labels=rna_orphan_columns, 
    column_labels=rna_samples, title="RNA Orphan reads", ylabel="Read count (in millions)",
    legend_title="Filter", yaxis_in_millions=True)




#!/usr/bin/env python

# This script will create a table of reads from humann2 log files. The table will have 
# total reads, unaligned after nucleotide search, and unaligned after translated search.
# The table will also include the total species number.

import sys
import os
import argparse

TOTAL_COUNT_TAG="reads; of these:\n"
NUCLEOTIDE_COUNT_TAG="Unaligned reads after nucleotide alignment:"
TRANSLATED_COUNT_TAG="Unaligned reads after translated alignment:"
SPECIES_COUNT_TAG="Total species selected from prescreen:"

def parse_arguments(args):
    """ 
    Parse the arguments from the user
    """
    parser = argparse.ArgumentParser(
        description= "Reads the HUMAnN2 logs and prints a table of read counts\n",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "-i", "--input", 
        help="the folder of log files\n[REQUIRED]", 
        metavar="<input_folder>", 
        required=True)
    parser.add_argument(
        "-o", "--output", 
        help="file to write counts table\n[REQUIRED]", 
        metavar="<output>", 
        required=True)

    return parser.parse_args()

def main():
    # parse arguments from the user
    args = parse_arguments(sys)

    # search subfolders to find the log files
    log_files=[]
    for path, directories, files in os.walk(args.input):
        for file in filter(lambda x: x.endswith(".log"), files):
            log_files.append(os.path.join(path,file))

    try:
        file_handle=open(args.output,"w")
    except EnvironmentError:
        sys.exit("Error: Unable to open output file: " + args.output)

    # write out the header
    file_handle.write("\t".join(["# samples","total reads","total nucleotide aligned","total translated aligned","total species"])+"\n")

    for file in log_files:
        sample=os.path.basename(file).split(".log")[0]
        data=[sample,"NA","NA","NA","NA"]
        for line in open(file):
            if line.endswith(TOTAL_COUNT_TAG):
                data[1]=int(line.split()[7])
            elif NUCLEOTIDE_COUNT_TAG in line:
                try:
                    data[2]=int(data[1]*((100-float(line.split()[-2]))/100.0))
                except (ValueError, TypeError):
                    print("Warning: Unable to compute nucleotide reads aligned from log line: " + line)
            elif TRANSLATED_COUNT_TAG in line:
                try:
                    data[3]=int(data[1]*((100-float(line.split()[-2]))/100.0))
                except (ValueError, TypeError):
                    print("Warning: Unable to compute translated reads aligned from log line: " + line)
            elif SPECIES_COUNT_TAG in line:
                data[4]=line.split()[-1]

        file_handle.write("\t".join([str(i) for i in data])+"\n")

    file_handle.close()
    print("Read table written to file: " + args.output)

if __name__ == "__main__":
    main()




#!/usr/bin/env python

""" This script will create otu tables from usearch alignments """

import sys

try:
    import argparse
except ImportError:
    sys.exit("Please upgrade to at least python v2.7")

import os
import re
from operator import add

USEARCH_HIT="H"
USEARCH_HIT_INDEX=0
USEARCH_QUERY_INDEX=8
USEARCH_TARGET_INDEX=9
SAMPLE_READ_DELIMITER="."
UNNAMED_TAXONOMY="Unclassified"
FASTA_SEQ_START=">"

def catch_open(file,write=None):
    """ Try to open the file, return file handle, or exit if unable to open """
    try:
        if write:
            file_handle=open(file,"w")
        else:
            file_handle=open(file)
    except EnvironmentError:
        sys.exit("ERROR: Unable to open file: " + file)
        
    return file_handle

def get_sample_id(sample):
    """ Get the sample id from a string with the read number """
    return SAMPLE_READ_DELIMITER.join(sample.split(SAMPLE_READ_DELIMITER)[0:-1])

def create_fasta_files(targets, green_genes_fasta, nonchimera_fasta, output_open_ref_fasta, output_closed_ref_fasta):
    """ Created open and closed reference fasta files with the sequence ids matching those in the otu tables """

    # try to open the output files
    file_handle_write_open_ref=catch_open(output_open_ref_fasta,write=True)
    file_handle_write_closed_ref=catch_open(output_closed_ref_fasta,write=True)

    # read through the nonchimera fasta file, writing to the output files
    write_seq=False
    with catch_open(nonchimera_fasta) as file_handle_read:
        for line in file_handle_read:
            if line[0] == FASTA_SEQ_START:
                # check if this sequence should be written to either output file
                id=line.rstrip().replace(FASTA_SEQ_START,"")
                if id in targets:
                    # write this to the open reference file
                    file_handle_write_open_ref.write(line)
                    write_seq=True
                else:
                    write_seq=False
            else:
                if write_seq:
                    file_handle_write_open_ref.write(line)

    # read through the green genes fasta file, writing to the output files
    write_seq=False
    with catch_open(green_genes_fasta) as file_handle_read:
        for line in file_handle_read:
            if line[0] == FASTA_SEQ_START:
                # check if this sequence should be written to either output file
                id=line.rstrip().replace(FASTA_SEQ_START,"")
                if id in targets:
                    # write this to the open and closed reference files
                    file_handle_write_open_ref.write(line)
                    file_handle_write_closed_ref.write(line)
                    write_seq=True
                else:
                    write_seq=False
            else:
                if write_seq:
                    file_handle_write_open_ref.write(line)
                    file_handle_write_closed_ref.write(line)

def create_otu_table(taxonomy_file, samples, denovo_otu_table, green_genes_uc, out_tsv, filtered_out_tsv, reads_to_otus):
    """ Create open and closed reference otu tables """

    # read in the green genes taxonomy ids to names file
    green_genes_taxonomy_ids={}
    for otu_id, taxonomy_name in read_tab_file(taxonomy_file):
        green_genes_taxonomy_ids[otu_id] = taxonomy_name

    otu_table = {}
    targets = set()
    known_otus = set()
    for alignment_type, otu_id, target in read_uc_file(green_genes_uc, all_alignments=True):
        if otu_id in denovo_otu_table:
            hits = denovo_otu_table[otu_id]
            # check if this is an alignment hit
            if alignment_type == USEARCH_HIT:
                if target in green_genes_taxonomy_ids:
                    taxonomy_name = green_genes_taxonomy_ids[target]
                    known_otus.add(otu_id)
                else:
                    sys.exit("ERROR: Alignment to green genes id not included in taxonomy file:" + target)
            else:
                taxonomy_name = UNNAMED_TAXONOMY
                target = otu_id

            # store the otu mapping to target
            targets.add(target)

            # record hits by the otu id and also the taxonomy name
            full_id = (target, taxonomy_name)
            if not full_id in otu_table:
                otu_table[full_id] = hits
            else:
                # add the current hits to those already recorded
                otu_table[full_id] = list(map(add, otu_table[full_id], hits))

    # write the two output files
    with catch_open(out_tsv, write=True) as file_handle:
        with catch_open(filtered_out_tsv, write=True) as filtered_file_handle:
            header="\t".join(["# OTU"]+samples+["taxonomy"])+"\n"
            file_handle.write(header)
            filtered_file_handle.write(header)
            for (taxonomy_id, taxonomy_name), hits in otu_table.items():
                output_line="\t".join([taxonomy_id]+list(map(str, hits))+[taxonomy_name])+"\n"
                file_handle.write(output_line)
                # do not write unclassified output to filtered file
                if not taxonomy_name == UNNAMED_TAXONOMY:
                    filtered_file_handle.write(output_line) 

    # compute the counts of reads mapping to known (included in green genes) and unknown otus 
    sample_known_read_counts={sample:0 for sample in samples}
    sample_unclassified_read_counts={sample:0 for sample in samples}
    for read, otus in reads_to_otus.items():
        sample = get_sample_id(read)
        # check if any of the otus are known
        if known_otus.intersection(otus):
            sample_known_read_counts[sample]+=1
        else:
            sample_unclassified_read_counts[sample]+=1

    return targets, sample_known_read_counts, sample_unclassified_read_counts

def read_tab_file(file):
    """ Read a tab delimited file """
    with catch_open(file) as file_handle:
        for line in file_handle:
            yield line.rstrip().split("\t")
    
def read_uc_file(uc_file, all_alignments=None):
    """ Read the uc file and return the type, query, and target """
    for data in read_tab_file(uc_file):
        try:
            query = data[USEARCH_QUERY_INDEX]
        except IndexError:
            query = ""
                  
        try:
            target = data[USEARCH_TARGET_INDEX]
        except IndexError:
            target = ""
                
        try:
            type = data[USEARCH_HIT_INDEX]
        except IndexError:
            type = ""
                
        if type and query and target:
            if all_alignments:
                yield type, query, target
            elif type == USEARCH_HIT:
                yield query, target
    
def write_denovo_otu_table(nonchimera_uc_mapping, output_file):
    """ Read the chimera uc mapping file and write a denovo table """
    
    # read the uc mapping file
    table = {}
    samples = set()
    queries_to_otus = {}
    for query, otu in read_uc_file(nonchimera_uc_mapping):
        sample = get_sample_id(query)
        samples.add(sample)
        if not otu in table:
            table[otu]={}
        table[otu][sample] = table[otu].get(sample,0) + 1
        if not query in queries_to_otus:
            queries_to_otus[query]=set()
        queries_to_otus[query].add(otu)
    
    # write the denovo output file
    sample_list=list(samples)
    denovo_otu_table={}
    with catch_open(output_file, write=True) as file_handle:
        # write the header
        file_handle.write("\t".join(["# OTU"]+list(map(str,sample_list)))+"\n")
        for otu, hits_to_otu in table.items():
            hits_by_sample = [hits_to_otu.get(sample, 0) for sample in sample_list]
            file_handle.write("\t".join([otu]+list(map(str,hits_by_sample)))+"\n")
            denovo_otu_table[otu]=hits_by_sample
            
    return sample_list, denovo_otu_table, queries_to_otus
                    
def count_reads_per_sample(file):
    """ Count the reads for each sample from the original fasta file """

    samples={}
    for line in catch_open(file):
        if line.startswith(">"):
            try:
                sample, read = line.replace(">","").split(SAMPLE_READ_DELIMITER)
            except ValueError:
                print("Warning: Sequence id has unexpected format, not included in total read count: " + line)
                continue
            if not sample in samples:
                samples[sample]=set()
            samples[sample].add(read)

    sample_counts={sample:len(list(reads)) for sample, reads in samples.items()}

    return sample_counts

def write_read_count_table(output_file, original_counts, known_counts, unknown_counts):
    """ Write a table of read counts per sample """

    samples=original_counts.keys()
    with catch_open(output_file, write=True) as file_handle:
        file_handle.write("\t".join(["# sample","original read count","reads mapping to OTU with taxonomy","reads mapping to unclassifed OTU"])+"\n")
        for sample in samples:
            file_handle.write("\t".join([sample]+[str(i) for i in [original_counts[sample],known_counts.get(sample,0),unknown_counts.get(sample,0)]])+"\n")

def parse_arguments(args):
    """ Parse the arguments from the user"""
    
    parser=argparse.ArgumentParser(description="Process usearch output to get taxonomy.")
    parser.add_argument('input_taxonomy',help="Taxonomy file from green genes.")
    parser.add_argument('input_greengenes_fasta',help="Fasta file from green genes.")
    parser.add_argument('input_greengenes_uc',help="The uc file of the alignment results from usearch to green genes.")
    parser.add_argument('input_nonchimera_uc',help="The uc file of alignment results from usearch to nonchimeras.")
    parser.add_argument('input_nonchimera_fasta',help="The uc nonchimeras fasta file.")
    parser.add_argument('input_original_fasta',help="The original fasta file containing all reads for all samples.")
    parser.add_argument('output_open_ref_tsv',help="The open ref otu table written.")
    parser.add_argument('output_open_ref_fasta',help="The open ref fasta written.")
    parser.add_argument('output_closed_ref_tsv',help="The closed reference otu table written.")
    parser.add_argument('output_closed_ref_fasta',help="The closed reference fasta written.")
    parser.add_argument('output_denovo_otu',help="The denovo otu table written.")
    parser.add_argument('output_read_count',help="The read count table written.")   
 
    return parser.parse_args()

def check_file_nonempty(file):
    """ Check the file is non-empty """
    
    try:
        size = os.stat(file).st_size
    except EnvironmentError:
        sys.exit("ERROR: Unable to access input file: " + file)
        
    if size == 0:
        sys.exit("ERROR: Input file is empty: " + file)
        
def main():
    # parse arguments
    args = parse_arguments(sys.argv)
    
    # check the input files are non-empty
    check_file_nonempty(args.input_nonchimera_uc)
    check_file_nonempty(args.input_greengenes_uc)

    # get the original read counts
    original_read_counts = count_reads_per_sample(args.input_original_fasta)

    samples, denovo_otu_table, reads_to_otus = write_denovo_otu_table(args.input_nonchimera_uc,args.output_denovo_otu)
    targets, sample_known_read_counts, sample_unknown_read_counts=create_otu_table(args.input_taxonomy,
        samples,denovo_otu_table,args.input_greengenes_uc,args.output_open_ref_tsv,args.output_closed_ref_tsv, reads_to_otus)
    create_fasta_files(targets,args.input_greengenes_fasta,args.input_nonchimera_fasta,args.output_open_ref_fasta,
        args.output_closed_ref_fasta)

    write_read_count_table(args.output_read_count, original_read_counts, sample_known_read_counts, sample_unknown_read_counts)

if __name__ == "__main__":
    main()

    
    



#!/usr/bin/env python

"""
bioBakery Workflows : A collection of AnADAMA2 workflows

Copyright (c) 2016 Harvard School of Public Health

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import sys

try:
    import argparse
except ImportError:
    sys.exit("Please upgrade to python v2.7")

import os
import subprocess

VERSION = "0.15.1"
WORKFLOW_FOLDER="workflows"
WORKFLOW_EXTENSION=".py"

def find_workflows():
    """ Search for installed workflows """
    
    workflow_folder=os.path.join(os.path.dirname(os.path.abspath(__file__)),WORKFLOW_FOLDER)
    workflows={}
    for file in os.listdir(workflow_folder):
        # look for files with the expected extension
        if file.endswith(WORKFLOW_EXTENSION):
            # do not need to add full path as these are also installed as executable scripts
            workflows[file.replace(WORKFLOW_EXTENSION,"")]=file
    
    return workflows

def parse_arguments(args,workflows):
    """ 
    Parse the arguments from the user
    """
    parser = argparse.ArgumentParser(
        description= "bioBakery workflows: A collection of AnADAMA2 workflows\n",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        prog="biobakery_workflows")
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s v"+VERSION)
    parser.add_argument(
        "workflow",
        choices=workflows,
        help="workflow to run")
    
    return parser.parse_args(args)

def run_workflow(args, workflow):
    """ Run the workflow with the arguments provided """
    
    try:
        command = [workflow]+args[2:]
        subprocess.call(command)
    except ( subprocess.CalledProcessError, EnvironmentError):
        sys.exit("Error: Unable to run workflow: " +" ".join(command))


def main():
    # find workflows
    workflows=find_workflows()
    
    # parse the arguments (only the first two as the rest are for the workflow)
    args=parse_arguments(sys.argv[1:2],workflows.keys())
    
    # run the workflow (providing all of the arguments)
    run_workflow(sys.argv,workflows[args.workflow])
    
    
if __name__ == "__main__":
    main()
    



#!/usr/bin/env python

import sys
import os
import subprocess
import string
import argparse

# rename the fastq files based on sample name instead of barcode
# to run: $ rename_fastq_files.py --input input_folder --input-barcodes barcodes.txt --output output_folder

def parse_arguments(args):
    """
    Parse the arguments from the user
    """
    parser = argparse.ArgumentParser(
        description= "Rename files from barcodes to sample names\n",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "-i", "--input",
        help="the input folder containing the fastq files\n[REQUIRED]",
        metavar="<input_folder>",
        required=True)
    parser.add_argument(
        "-b", "--input-barcodes",
        help="the tab-delimited file of sample names and barcodes\n[REQUIRED]",
        metavar="<barcodes.tsv>",
        required=True)
    parser.add_argument(
        "-o", "--output",
        help="the output folder to write the new fastq files\n[REQUIRED]",
        metavar="<output_folder>",
        required=True)
    parser.add_argument(
        "--input-basename1",
        help="the basename of the input file with SEQ replacing the barcode for R1",
        default="1_ABCD.1.SEQ.unmapped.1.fastq.gz")
    parser.add_argument(
        "--input-basename2",
        help="the basename of the input file with SEQ replacing the barcode for R2",
        default="1_ABCD.1.SEQ.unmapped.2.fastq.gz")
    parser.add_argument(
        "--output-basename1",
        help="the basename of the output file with SAMPLE to be replaced with the sample name for R1",
        default="SAMPLE.R1.fastq.gz")
    parser.add_argument(
        "--output-basename2",
        help="the basename of the output file with SAMPLE to be replaced with the sample name for R2",
        default="SAMPLE.R2.fastq.gz")

    return parser.parse_args()


def main():

    args=parse_arguments(sys)

    # get the set of barcodes (and reverse complements) for each sample
    revcmp=string.maketrans("ATGC","TACG")
    barcodes={}
    for line in open(args.input_barcodes):
        data=line.rstrip().split("\t")
        newseq=data[1].translate(revcmp)[::-1]
        barcodes[newseq]=data[0]

    # create the output folder if it does not exist
    if not os.path.isdir(args.output):
        os.makedirs(args.output)

    # rename files, copying to new output folder
    not_found=0
    for seq, sample in barcodes.items():
        command1=["cp",os.path.join(args.input,args.input_basename1.replace("SEQ",seq)),
            os.path.join(args.output,args.output_basename1.replace("SAMPLE",sample))]
        command2=["cp",os.path.join(args.input,args.input_basename2.replace("SEQ",seq)),
            os.path.join(args.output,args.output_basename2.replace("SAMPLE",sample))]
        try:
            print(command1)
            subprocess.check_output(command1)
            print(command2)
            subprocess.check_output(command2)
        except subprocess.CalledProcessError:
            print("Unable to find sample: {}".format(sample))
            not_found+=1

    print("Total not found: {}".format(not_found))

if __name__ == "__main__":
    main()




#!/usr/bin/env python

"""
bioBakery Workflows: 16S workflow for visualization

Copyright (c) 2017 Harvard School of Public Health

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

# import the workflow class from anadama2
from anadama2 import Workflow

# import the document templates from biobakery_workflows
from biobakery_workflows import document_templates, utilities

# import the files for descriptions and paths
from biobakery_workflows import files
import os

# create a workflow instance, providing the version number and description
# remove the input folder option as it will be replaced with multiple input files
workflow = Workflow(version="0.1", remove_options=["input"],
                    description="A workflow for 16S visualization")
                    
# add the custom arguments to the workflow 
# create a custom description for the input argument listing all expected input files
input_desc="A folder containing the final products from the 16s data workflow.\n\nThe input folder must include the following:\n\n"  
                      
workflow.add_argument("input",desc=input_desc,required=True)

# add the custom arguments to the workflow
workflow.add_argument("project-name",desc="the name of the project",required=True)
workflow.add_argument("input-metadata",desc="the metadata file (samples as columns or rows)")
workflow.add_argument("input-picard",desc="the folder of picard quality score files")
workflow.add_argument("input-picard-extension",desc="the extensions for the picard quality score files", default="quality_by_cycle_metrics")
workflow.add_argument("metadata-categorical",desc="the categorical features", action="append", default=[])
workflow.add_argument("metadata-continuous",desc="the continuous features", action="append", default=[])
workflow.add_argument("metadata-exclude",desc="the features to exclude", action="append", default=[])
workflow.add_argument("exclude-workflow-info",desc="do not include data processing task info in report", action="store_true")
workflow.add_argument("format",desc="the format for the report", default="pdf", choices=["pdf","html"])

# get the arguments from the command line
args = workflow.parse_args()

otu_table = files.SixteenS.path("otu_table_closed_reference",args.input, error_if_not_found=True)

# read and label the metadata
metadata=None
metadata_labels=None
if args.input_metadata:
    metadata=utilities.read_metadata(args.input_metadata, otu_table, ignore_features=args.metadata_exclude, otu_table=True)
    metadata_labels, metadata=utilities.label_metadata(metadata, categorical=args.metadata_categorical, continuous=args.metadata_continuous)

templates=[utilities.get_package_file("16S")]

log_file=None
# add the template for the data processing information
if not args.exclude_workflow_info:
    log_file=files.Workflow.path("log", args.input, error_if_not_found=True)

# identify method and list the required and optional files for the workflow
# these are expected to be included in the input folder

# for dada2/its workflow
if os.path.isfile(files.SixteenS.path("error_ratesF", args.input, error_if_not_found=False)):
    method = "dada2"
    if os.path.isdir(files.SixteenS.path("filtN", args.input, error_if_not_found=False)):
        method = "its"
    doc_title = method.upper() + " 16s Report"
    input_files = {
        "required": [
            "otu_table_closed_reference",
            "msa_nonchimera",
            "counts_each_step",
            "error_ratesF",
            "error_ratesR",
            "readF_qc",
            "readR_qc"]}

    # get the paths for the required files and check they are found
    centroid_fasta = files.SixteenS.path("msa_nonchimera", args.input, error_if_not_found=True)
    counts_each_step = files.SixteenS.path("counts_each_step", args.input, error_if_not_found=True)
    error_ratesF = files.SixteenS.path("error_ratesF", args.input, error_if_not_found=True)
    error_ratesR = files.SixteenS.path("error_ratesR", args.input, error_if_not_found=True)
    readF_qc = files.SixteenS.path("readF_qc", args.input, error_if_not_found=True)
    readR_qc = files.SixteenS.path("readR_qc", args.input, error_if_not_found=True)

    methoddepends = [counts_each_step, otu_table, centroid_fasta, error_ratesF, error_ratesR, readF_qc, readR_qc]

    # variables
    methodvars = {
        "title": doc_title,
        "project": args.project_name,
        "method": method,
        "outputdir": args.output,
        "otu_table": otu_table,
        "counts_each_step": counts_each_step,
        "error_ratesF": error_ratesF,
        "error_ratesR": error_ratesR,
        "readF_qc": readF_qc,
        "readR_qc": readR_qc,
        "format": args.format,
        "log": log_file,
        "metadata": metadata,
        "metadata_labels": metadata_labels,
        "picard": args.input_picard,
        "picard_ext": args.input_picard_extension}
 
# for usearch workflow
else:
    method = "usearch"
    input_files={
	 "required":[
	 	"otu_table_closed_reference",
		"otu_table_open_reference",
		"read_count_table",
		"eestats2"]}

    # get the paths for the required files and check they are found
    otu_open_table=files.SixteenS.path("otu_table_open_reference",args.input, error_if_not_found=True)
    read_count_table=files.SixteenS.path("read_count_table",args.input, error_if_not_found=True)
    eestats_table=files.SixteenS.path("eestats2",args.input, error_if_not_found=True)
    centroid_fasta = files.SixteenS.path("msa_nonchimera",args.input, none_if_not_found=True)
    centroid_closed_fasta = files.SixteenS.path("msa_closed_reference", args.input, none_if_not_found=True)
	       
    methoddepends=[otu_table, read_count_table, otu_open_table, eestats_table, centroid_fasta, centroid_closed_fasta]
    
    # variables
    methodvars={"title":"USEARCH 16S Report",
          "project":args.project_name,
          "method":method,
          "otu_table":otu_table,
          "read_count_table":read_count_table,
          "eestats_table":eestats_table,
          "format":args.format,
          "log":log_file,
          "metadata":metadata,
          "metadata_labels":metadata_labels,
          "picard":args.input_picard,
          "picard_ext":args.input_picard_extension}


# listing all expected input files
input_desc+=files.SixteenS.list_file_path_description("",input_files)


if not args.exclude_workflow_info:
    templates += [utilities.get_package_file("workflow_info")]


# add the document to the workflow
doc_task=workflow.add_document(
    templates=templates,
    depends=methoddepends, 
    targets=workflow.name_output_files("16S_report."+args.format),
    vars=methodvars,
    table_of_contents=True)

# add an archive of the document and figures, removing the log file
# the archive will have the same name and location as the output folder
workflow.add_archive(
    depends=[args.output,doc_task],
    targets=args.output+".zip",
    remove_log=True)

# start the workflow
workflow.go()



# This script will create a new folder of symlinked files with the new id naming convention using the mapping file.
# $ python rename_files_to_sample_ids.py infolder outfolder mapping file

import os
import sys
import subprocess

infolder=os.path.abspath(sys.argv[1])
outfolder=os.path.abspath(sys.argv[2])

mapping_file=sys.argv[3]

# read in all of the sample ids
ids={}
with open(mapping_file) as file_handle:
    header = file_handle.readline()
    for line in file_handle:
        data = line.rstrip().split("\t")
        ids[data[1]]=data[0]

mapped_ids=set()
unmapped_ids=0
for filename in os.listdir(infolder):
    if not "contam" in filename and not filename.endswith(".log"):
        basename = filename.split(".")[0].split("_")[0]
        if not basename in ids:
            print "Unable to find basename in mapping file: "+basename
            unmapped_ids+=1
        else:
            mapped_ids.add(basename)
            newfilename=ids[basename]+".fastq"
            cmmd=["ln","-s",os.path.join(infolder,filename),os.path.join(outfolder,newfilename)]
            #print cmmd
            subprocess.check_call(cmmd)

print "Total mapped ids: "+ str(len(list(mapped_ids)))
print "Total unmapped ids: "+ str(unmapped_ids)



#!/usr/bin/env python

import sys
import re
import argparse

from biobakery_workflows import utilities

# This script will check a fastq file is of the expected format.
# To run: $ check_fastq_format.py --input input.fastq

def parse_arguments(args):
    # Parse the arguments from the user
    
    parser = argparse.ArgumentParser(
        description= "Check fastq format\n",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "-i", "--input",
        help="the fastq file\n[REQUIRED]",
        metavar="<input.fastq>",
        required=True)
    parser.add_argument(
        "-s","--stop-first-error",
        help="stop after the first formatting error\n[DEFAULT: print all formatting errors]",
        action='store_true')

    return parser.parse_args() 

def write_error(message,id,sequence,qual_id,quality_scores,stop,errors_found):
    # write the error message including the full sequence 
    print("FORMAT ERROR: " + message)
    print("".join([id,sequence,qual_id,quality_scores]))
    if stop:
        sys.exit()

    return errors_found+1

def main():
    # parse command line arguments
    args=parse_arguments(sys)

    # this is a regex of the allowable quality score ascii
    regex = re.compile(r"[a-zA-Z0-9#;/>@?<=\-\\&,\'\$%\*\+\(\)]")

    # check for possible fastq formatting errors
    errors_found=0
    for lines in utilities.read_file_n_lines(args.input,4):
        seq_id, sequence, qual_id, quality_scores = lines
        remainder = regex.sub("",re.escape(quality_scores.rstrip()))
        if len(remainder) > 0:
            errors_found=write_error("Unexpected character in quality scores (found: {})".format(remainder),
                seq_id,sequence, qual_id, quality_scores, args.stop_first_error, errors_found)
        if len(sequence) != len(quality_scores):
            errors_found=write_error("Sequence and quality score lengths differ",
                seq_id,sequence, qual_id, quality_scores, args.stop_first_error, errors_found)
        if len(quality_scores.rstrip()) == 0 or len(sequence.rstrip()) == 0:
            errors_found=write_error("Empty sequence", seq_id, sequence, 
                qual_id, quality_scores, args.stop_first_error, errors_found)
        if seq_id[0] != "@":
            errors_found=write_error("Unexpected format in sequence id (first character is not @)",
                seq_id,sequence, qual_id, quality_scores, args.stop_first_error, errors_found)
        if qual_id[0] != "+":
            errors_found=write_error("Unexpected format in quality id (first character is not +)",
                seq_id,sequence, qual_id, quality_scores, args.stop_first_error, errors_found)

    if errors_found == 0:
        print("No errors identified in fastq format")
    else:
        print("{} errors identified in fastq format".format(errors_found))

if __name__ == "__main__":
    main()




"""
bioBakery Workflows setup

To run: python setup.py install
"""

import os
import sys

try:
    import setuptools
except ImportError:
    sys.exit("Please install setuptools.")

# try to import urllib.request.urlretrieve for python3
try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve
    
from glob import glob    

VERSION = "0.15.1"

AUTHOR = "bioBakery workflows development team"
AUTHOR_EMAIL = "biobakery-users@googlegroups.com"
MAINTAINER = "Lauren McIver"
MAINTAINER_EMAIL = "lauren.j.mciver@gmail.com"

setuptools.setup(
    name="biobakery_workflows",
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    version=VERSION,
    license="MIT",
    description="bioBakery workflows: A collection of meta'omic analysis workflows",
    long_description="bioBakery workflows is a collection of workflows and tasks for "+\
        "executing common microbial community analyses using standardized, validated "+\
        "tools and parameters. Quality control and statistical summary reports are "+\
        "automatically generated for most data types, which include 16S amplicons, "+\
        "metagenomes, and metatranscriptomes. Workflows are run directly from the "+\
        "command line and tasks can be imported to create your own custom workflows. "+\
        "The workflows and tasks are built with AnADAMA2 which allows for parallel "+\
        "task execution locally and in a grid compute environment.",
    url="http://huttenhower.sph.harvard.edu/biobakery_workflows",
    keywords=['microbial','microbiome','bioinformatics','microbiology','metagenomic','metatranscriptomic','anadama2','humann2','metaphlan2','strainphlan'],
    platforms=['Linux','MacOS'],
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Operating System :: MacOS",
        "Operating System :: Unix",
        "Programming Language :: Python :: 2.7",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
        ],
    install_requires=['anadama2>=0.5.2'],
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            'biobakery_workflows = biobakery_workflows.biobakery_workflows:main',
            'biobakery_workflows_databases = biobakery_workflows.biobakery_workflows_databases:main',
        ]},
    package_data={
        'biobakery_workflows' : [
            'workflows/*.py',
            'document_templates/*.py',
            'document_templates/*.mdw',
            'data/*',
            'Rscripts/*'
        ]},
    data_files=[
        ("tutorial/input/", glob("examples/tutorial/input/*")),
        ("tutorial/kneaddata_demo_db/Homo_sapiens_demo/", glob("examples/tutorial/kneaddata_demo_db/Homo_sapiens_demo/*")),
        ("tutorial/kneaddata_demo_db/SILVA_demo/", glob("examples/tutorial/kneaddata_demo_db/SILVA_demo/*"))
    ],
    scripts=glob('biobakery_workflows/workflows/*py')+glob('biobakery_workflows/scripts/*py'),
    test_suite='tests.get_test_suite',
    zip_safe = False
 )



#!/usr/bin/env python

import sys
import os
import argparse
import subprocess
import random

try:
    from humann2 import config
    from humann2 import store
except ImportError:
    sys.exit("Please install humann2.")

# This script will take as input the bowtie2 alignment file from
# running humann2 along with a list of species and pathways. The selection list
# will be formatted as two tab delimited columns of species \t pathway. 
# To include all pathways for a species use the key "ALL". It can
# have headers or comments starting with "#". The script will output a 
# reduced fasta (or fastq if selected) file that only includes reads that 
# will map to those species and pathways when running the fastq file as input to humann2.

# Please note since some pathways have reactions that overlap with other pathways
# more than just the selected pathways may appear for a specific species. 

# If running humann2 with the output from the script does not identify any species
# in the metaphlan2 step, recreate the humann2 input file with this script adding
# the option "--add-unintegrated". This option will add reads that map to a species
# in the humann2 nucleotide alignment step but the gene families associated with these
# alignments are not included in any pathways (and are counted as "unintegrated" in 
# the humann2 pathway abundance output file).

def parse_arguments(args):
    """ 
    Parse the arguments from the user
    """
    parser = argparse.ArgumentParser(
        description= "Create demo file\n",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "--input-sam",
        help="the sam bowtie2 alignment temp file from running HUMAnN2\n[REQUIRED]",
        metavar="<input.sam>",
        required=True)
    parser.add_argument(
        "--input-selection",
        help="the species and pathways requested\n[REQUIRED]",
        metavar="<input.tsv>",
        required=True)
    parser.add_argument(
        "--add-unintegrated",
        help="add reads that do not map to pathways for each selected species\n[OPTIONAL]",
        action="store_true")
    parser.add_argument(
        "--output",
        help="file to write the sub-sampled file\n[REQUIRED]",
        metavar="<output.fastq>",
        required=True)
    parser.add_argument(
        "--output-format",
        help="the format to write the output file\n[OPTIONAL]",
        choices=["fastq","fasta"],
        default="fasta")
    parser.add_argument(
        "--gene-families",
        help="the gene families to select with\n[OPTIONAL]",
        choices=["UniRef50","UniRef90"],
        default="UniRef90")
    parser.add_argument(
        "--percent",
        help="the percent of reads to output (subsampled by reaction set for requested pathways)\n[OPTIONAL]",
        type=int,
        default=100)
    parser.add_argument(
        "--add-markers",
        help="add in the reads that align to markers for the selected species using MetaPhlAn2\n[OPTIONAL]",
        action="store_true")
    parser.add_argument(
        "--min-markers",
        help="the minimum number of reads that map to markers for each species using MetaPhlAn2\n[OPTIONAL]",
        type=int,
        default=20)
    parser.add_argument(
        "--min-reads-per-marker",
        help="the minimum number of reads per marker for each species using MetaPhlAn2\n[OPTIONAL]",
        type=int,
        default=5)
    parser.add_argument(
        "--add-trimmable",
        help="add reads that would be trimmed based on length\n[OPTIONAL]",
        action="store_true")

    return parser.parse_args()

def write_sequence(file_handle, read_name, sequence, quality_scores, output_format):
    """ Write the sequence to the output file in the selected format """
    
    if output_format == "fasta":
        file_handle.write(">"+read_name+"\n")
        file_handle.write(sequence+"\n")
    else:
        file_handle.write("@"+read_name+"\n")
        file_handle.write(sequence+"\n")
        file_handle.write("+\n")
        file_handle.write(quality_scores+"\n")

def read_sam(file, gene_families_to_use):
    """ Read the sam file, only reading alignments """

    for line in open(file):
        data=line.rstrip().split("\t")
        if len(data) > config.sam_read_quality:
            read_name=data[config.sam_read_name_index]
            reference_name=data[config.sam_reference_index]
            sequence=data[config.sam_read_index]
            quality_scores=data[config.sam_read_quality]

            reference_data = reference_name.split(config.chocophlan_delimiter)
            try:
                uniref50=reference_data[-2]
                uniref90=reference_data[-3]
                species=reference_data[-4].split(".")[-1]
            except IndexError:
                # ignore reads that do not map to the reference
                continue
            
            gene_family=uniref90
            if gene_families_to_use == "UniRef50":
                gene_family=uniref50

            yield species, gene_family, read_name, sequence, quality_scores

def main():

    args=parse_arguments(sys)
    
    # read in the gene families to reactions database
    print("Reading gene families to reactions database")
    reactions_database=store.ReactionsDatabase(config.pathways_database_part1)
    
    # read in the reactions to pathways database
    print("Reading reactions to pathways database")
    pathways_database=store.PathwaysDatabase(config.pathways_database_part2, reactions_database)
    
    # read in the species and pathways selected
    print("Finding gene families for each species for the pathways selected")
    genefamilies={}
    reactions={}
    for line in open(args.input_selection):
        # remove starting and ending spaces
        line=line.strip()
        if not line.startswith("#"):
            try:
                species, pathway = line.split("\t")
            except IndexError:
                print("Warning: Skipping selection line because of format: " + line)
                continue
            if not species in genefamilies:
                genefamilies[species]=set()
                reactions[species]=set()
            # get the reactions and then gene families for the pathway
            if pathway.upper() == "ALL":
                genefamilies[species]="ALL"
                reactions[species]="ALL"
            else:
                for reaction in pathways_database.find_reactions(pathway):
                    reactions[species].add(reaction)
                    genefamilies[species].update(list(filter(lambda x: x.startswith(args.gene_families), reactions_database.find_genes(reaction))))
     
    # get the set of gene families in any pathway           
    all_genefamilies_in_pathways=set()
    for reaction in pathways_database.reaction_list():
        all_genefamilies_in_pathways.update(list(filter(lambda x: x.startswith(args.gene_families), reactions_database.find_genes(reaction))))

    # open the output file
    try:
        file_handle=open(args.output,"w")
    except EnvironmentError:
        sys.exit("ERROR: Unable to open output file: " + args.output)
                
    # find the reads for the species and pathways selected
    print("Reading sam file")
    selected_reads=set()
    reads_to_gene_familes={}
    reaction_totals={}
    
    # open a fasta file to write the species specific sequences for input to metaphlan2
    if args.add_markers:
        species_fasta_file=args.output+".species_specific_reads.fasta"
        species_fasta_file_handle=open(species_fasta_file,"w")
        
    selected_reads_to_species={}
    for species, gene_family, read_name, sequence, quality_scores in read_sam(args.input_sam, args.gene_families):
        # record the reads that align to species/gene families requested
        requested_genes_for_species=genefamilies.get(species,[])
        if (gene_family in requested_genes_for_species) or ("ALL" in requested_genes_for_species) or (args.add_unintegrated and not(gene_family in all_genefamilies_in_pathways)):
            selected_reads.add(read_name)
            selected_reads_to_species[species]=selected_reads_to_species.get(species,0)+1
            if not read_name in reads_to_gene_familes:
                reads_to_gene_familes[read_name]=set()
            
            reads_to_gene_familes[read_name].add(gene_family)
            
            # add to the reaction totals count
            for reaction in reactions_database.find_reactions(gene_family):
                reaction_totals[reaction]=reaction_totals.get(reaction,0)+1.0
                
        # if this is a species in the list, and we are adding markers, write this to the fasta file for input to metaphlan2
        if args.add_markers and species in genefamilies.keys():
            write_sequence(species_fasta_file_handle, read_name, sequence, quality_scores, "fasta")
           
    print("Total reads per species based on gene families")
    for species, count in selected_reads_to_species.items():
        print(species+"\t"+str(count))         

    # close the species fasta file handle
    if args.add_markers:
        species_fasta_file_handle.close()
            
    print("Total reads found: " + str(len(selected_reads)))
    
    # get the markers reads from the species fasta file
    marker_reads_to_add=set()
    read_to_species_marker={}
    read_to_marker_name={}
    all_reads_mapping_to_markers=set()
    if args.add_markers:
        print("Finding reads mapped to markers with MetaPhlAn2")
        
        print("Running MetaPhlAn2")
        metaphlan2_marker_file=species_fasta_file+".marker_alignments.tsv"
        metaphlan2_bowtie2_file=species_fasta_file+".bowtie2.tsv"
        
        try:
            # remove the bowtie2 output file if it exists to prevent metaphlan2 error
            os.remove(metaphlan2_bowtie2_file)
        except EnvironmentError:
            pass
        
        output=subprocess.check_output(["metaphlan2.py","--input_type","fasta",species_fasta_file,"-t","reads_map","-o",metaphlan2_marker_file,"--bowtie2out",metaphlan2_bowtie2_file])
        
        # read through the file to identify the maker reads to add
        for line in open(metaphlan2_marker_file):
            if not line.startswith("#"):
                read_name, taxon = line.rstrip().split("\t")
                species = taxon.split("|")[-1]
                if species in genefamilies.keys():
                    marker_reads_to_add.add(read_name)
                    if not species in read_to_species_marker:
                        read_to_species_marker[species]=set()
                        
                    read_to_species_marker[species].add(read_name)
                    all_reads_mapping_to_markers.add(read_name)
                    
        # read through the file to identify the name of the marker the reads map to
        for line in open(metaphlan2_bowtie2_file):
            try:
                read_name, marker_name = line.rstrip().split("\t")
            except IndexError:
                continue
            
            if read_name in all_reads_mapping_to_markers:
                read_to_marker_name[read_name]=marker_name
                    
        print("Found a total of "+str(len(marker_reads_to_add))+" reads to species markers")
    
    if args.percent < 100:
        print("Filtering reads by percent requested: " + str(args.percent))
        filtered_reads=set()
        current_reaction_totals={reaction:0 for reaction in reaction_totals.keys()}
        # go through the reads, adding until there is enough in the reaction list
        # to hit the percent requested
        for read_name in selected_reads:
            # check if this read is needed into increase the reaction percents
            add_read=False
            for gene in reads_to_gene_familes[read_name]:
                for reaction in reactions_database.find_reactions(gene):
                    if (current_reaction_totals[reaction]/reaction_totals[reaction])*100 < args.percent:
                        add_read=True
                        
            if add_read:
                filtered_reads.add(read_name)
                # update the current reaction counts
                for gene in reads_to_gene_familes[read_name]:
                    for reaction in reactions_database.find_reactions(gene):
                        current_reaction_totals[reaction]+=1
                        
        # update the selected reads to those that are filtered
        selected_reads=filtered_reads
        print("Total reads after filtering: "+str(len(selected_reads)))
    
    # check to make sure at least 200 marker reads (or command line setting) for each species are present
    # make sure there reads are spread over the markers for the sample
    if args.add_markers:
        # for each species, check the number of maker reads present
        print("Counting markers in set to determine if reads need to be added to meet min markers")
        for species, reads_for_species in read_to_species_marker.items():
            # count how many of the reads are in the selected reads list
            overlap=selected_reads.intersection(reads_for_species)
            total_overlap=len(list(overlap))
            if total_overlap < args.min_markers:
                # add in more reads to hit the min markers setting for this species
                max_reads_to_add=list(reads_for_species.difference(selected_reads))
                # group the reads based on the species markers
                to_add_by_markers={}
                for read_name in max_reads_to_add:
                    marker_for_read = read_to_marker_name[read_name]
                    if not marker_for_read in to_add_by_markers:
                        to_add_by_markers[marker_for_read]=set()
                    to_add_by_markers[marker_for_read].add(read_name)
                    
                selected_reads_by_markers={}
                for read_name in selected_reads:
                    try:
                        marker_for_read = read_to_marker_name[read_name]
                    except KeyError:
                        # ignore reads that do not map to markers
                        continue
                    if not marker_for_read in selected_reads_by_markers:
                        selected_reads_by_markers[marker_for_read]=set()
                    selected_reads_by_markers[marker_for_read].add(read_name)
                
                # count the total reads to add    
                total_to_add=args.min_markers - total_overlap
                total_added=0
                
                # get a list of all of the markers for the species
                all_species_markers=set(to_add_by_markers.keys())
                all_species_markers.update(selected_reads_by_markers.keys())
                
                # start with the markers not already included, adding one read to each
                for marker_name in all_species_markers.difference(selected_reads_by_markers.keys()):
                    total_markers_in_set=len(list(to_add_by_markers[marker_name]))
                    end_index = args.min_reads_per_marker if total_markers_in_set >= args.min_reads_per_marker else total_markers_in_set
                    
                    # add at most min reads per markers for this marker set to the set of selected reads
                    to_add=list(to_add_by_markers[marker_name])[:end_index]
                    selected_reads.update(to_add)
                    if not marker_name in selected_reads_by_markers:
                        selected_reads_by_markers[marker_name]=set()
                    selected_reads_by_markers[marker_name].update(to_add)
                    
                    # remove the added reads from the set of reads to add
                    to_add_by_markers[marker_name]=to_add_by_markers[marker_name].difference(to_add)
                    
                    # decrease the total amount to add
                    total_added+=end_index
                    
                    print("Adding " +str(end_index)+ " total reads to fill empty marker for species " + species)
                    
                # next add more reads for those sets that already have markers, starting with the smallest
                # set of markers to the largest
                marker_counts={marker:len(list(reads)) for marker, reads in selected_reads_by_markers.items()} 
                for marker_name in sorted(marker_counts, key=marker_counts.get):
                    try:
                        total_markers_in_set=len(list(to_add_by_markers[marker_name]))
                    except KeyError:
                        # ignore errors for markers that are not included in the to add list (only in the selected reads)
                        continue
                    
                    end_index = args.min_reads_per_marker if total_markers_in_set >= args.min_reads_per_marker else total_markers_in_set
                    
                    # add only so many reads to get to min reads per markers for this set
                    end_index=end_index-len(list(selected_reads_by_markers[marker_name]))
                    
                    # add at most min reads per marker reads for this marker set to the set of selected reads
                    if end_index > 0:
                        to_add=list(to_add_by_markers[marker_name])[:end_index]
                        selected_reads.update(to_add)
                        selected_reads_by_markers[marker_name].update(to_add)
                            
                        # remove the added reads from the set of reads to add
                        to_add_by_markers[marker_name]=to_add_by_markers[marker_name].difference(to_add)
                            
                        total_added+=end_index
                        print("Adding " +str(end_index)+ " total reads to fill slightly full marker for species " + species)
                    
                # add more reads to get to the min reads added value
                total_to_add=total_to_add-total_added if total_added < total_to_add else 0
                try:
                    selected_reads.update(max_reads_to_add[:total_to_add])
                    total_added+=total_to_add
                    print("Added "+str(total_to_add)+" reads for species "+species+" to meet min markers")
                except (TypeError, IndexError):
                    continue
                
            print("Total reads added for species "+species+" :"+str(total_added)) 
        print("Total reads after counting markers: " + str(len(selected_reads)))
   
    # determine how many trimmable reads to write
    total_trimmable = 0
    trimmable = []
    if args.add_trimmable:
        total_trimmable = int(random.uniform(50,100)) 
        print("Adding "+str(total_trimmable)+" total trimmable reads")
 
    print("Writing output file")
    for species, gene_family, read_name, sequence, quality_scores in read_sam(args.input_sam, args.gene_families):
        # write the read sequences requested once
        if read_name in selected_reads:
            write_sequence(file_handle, read_name, sequence, quality_scores, args.output_format)
            selected_reads.remove(read_name)
            if len(trimmable) < total_trimmable:
                trimmable.append([read_name, sequence, quality_scores])

    # write the trimmable reads
    for read_name, sequence, quality_scores in trimmable:
        new_length=int(len(sequence)/2.0)
        write_sequence(file_handle, "trimmable_"+read_name, sequence[0:new_length], quality_scores[0:new_length], args.output_format)

    print("Output file written: " + args.output)
    
if __name__ == "__main__":
    main()



import cPickle as pickle
import bz2

from anadama2 import Workflow

# To run
# $ python pull_out_reads_by_species_metaphlan2_results.py --input input_sam --output output_fastq
# This will look for the metaphlan2 sam output files named *_bowtie2.sam in the input folder and write
# files *_metaphlan2_marker_aligned_subset.fasta to the output folder (one for each input sam file).
# The fasta reads will be any of the sample reads that map to a marker associated with one of the 
# species in the "--species-list" file. This file should have one species per line and be formatted 
# with the metaphlan2 species naming convention. More specifically, the species file should list
# one per line with metaphlan2 format (ie "s__Gemella_sanguinis") and for unknown species
# include the genus in this file (ie "s__Gemella_unclassified" should be included in the file as "g__Gemella").
# The metaphlan2 pkl database is also required for this script to run and can be provided 
# with the option "--pkl-database". 

SAM_READ_NAME_INDEX = 0
SAM_REFERENCE_NAME_INDEX = 2
SAM_SEQ_INDEX = 9

workflow = Workflow()

# input folder should have sam alignment files from metaphlan2 run
workflow.add_argument("pkl-database", desc="MetaPhlAn2 pkl database", default="metaphlan2_db/mpa_v20_m200.pkl")
workflow.add_argument("species-list", desc="the list of species to pull reads for", default="species_list.txt")
workflow.add_argument("input-tag-extension", desc="the file name tag and extension", default="_bowtie2.sam")
args = workflow.parse_args()

def find_reads(task):
    # read in the species
    with open(args.species_list) as file_handle:
        species_list = [taxon.rstrip() for taxon in file_handle.readlines()]

    db = pickle.load(bz2.BZ2File(args.pkl_database, 'r'))

    marker_to_species={}
    for marker,info in db['markers'].items():
        if info['clade'] in species_list:
            marker_to_species[marker]=info['clade']

    # read in the sam file and pull out the reads that align with the markers
    with open(task.targets[0].name, "w") as file_handle_write:
        with open(task.depends[0].name) as file_handle:
            for line in file_handle:
                if not line.startswith("@"):
                    data=line.rstrip().split("\t")
                    reference=data[SAM_REFERENCE_NAME_INDEX]
                    if reference in marker_to_species.keys():
                        seq_id = ";".join([data[SAM_READ_NAME_INDEX],marker_to_species[reference]])
                        seq = data[SAM_SEQ_INDEX]
                        file_handle_write.write("\n".join([">"+seq_id,seq])+"\n")

# for each of the input files write the fasta file of reads
for infile in workflow.get_input_files(extension=args.input_tag_extension):
    outfile = workflow.name_output_files(infile).replace(args.input_tag_extension,"_metaphlan2_marker_aligned_subset.fasta")
    workflow.add_task(
        find_reads,
        depends=infile,
        targets=outfile)

workflow.go()




#!/usr/bin/env python

""" This script will take as input paired files of sequences which are already 
demultiplexed and split them into files for each sample name using the 
corresponding index file and barcodes file.
"""

import argparse
import sys
import os
import string

def parse_arguments(args):
    """ Parse the arguments from the user """
    
    parser = argparse.ArgumentParser(
        description= "Demultiplex files using barcodes and index (sequences no longer include barcodes)\n",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "--input-index", 
        help="input index file \n[REQUIRED if barcodes are not in read ids]")
    parser.add_argument(
        "--input-read1", 
        help="input read1 file \n[REQUIRED]", 
        required=True)
    parser.add_argument(
        "--input-read2", 
        help="input read2 file \n[REQUIRED]", 
        required=True)
    parser.add_argument(
        "--input-barcodes", 
        help="input barcodes file \n[REQUIRED]", 
        required=True)
    parser.add_argument(
        "--reverse-complement", 
        help="use reverse complements of index sequences\n", 
        action="store_true")
    parser.add_argument(
        "-o", "--output", 
        help="directory to write output files\n[REQUIRED]", 
        metavar="<output>", 
        required=True)
    
    return parser.parse_args()

def read_fastq_file(file):
    """ Read a fastq file (not compressed) """
    
    line_set=[]
    with open(file) as file_handle:
        for line in file_handle:
            if len(line_set) == 4:
                yield line_set
                line_set=[]
            line_set.append(line)
            
def read_barcode_file(file):
    """ Read through the barcode file storing samples and sequences """
    
    barcodes={}
    with open(file) as file_handle:
        header=file_handle.readline()
        for line in file_handle:
            data=line.rstrip().split("\t")
            # replace periods with underscore in sample ids
            data[0]=data[0].replace(".","_")
            if data[1] in barcodes and data[0] != barcodes[data[1]]:
                sys.exit("ERROR: Duplicate barcode with different sample match: "+ data[1])
            barcodes[data[1]]=data[0]
            
    return barcodes
            
def main():
    # parse arguments from the user
    args = parse_arguments(sys.argv)
    
    # if the output folder does not exist then create it
    if not os.path.isdir(args.output):
        try:
            os.makedirs(args.output)
        except EnvironmentError:
            sys.exit("ERROR: Unable to create output directory: " + args.output)
    
    # read through the barcode file
    barcodes=read_barcode_file(args.input_barcodes)
    
    # create rev cmp function
    rev_cmp=string.maketrans("ACGT","TGCA")
    
    # read through read1/2 files pulling a sequence set each time
    read1_lines = read_fastq_file(args.input_read1)
    read2_lines = read_fastq_file(args.input_read2)
    
    # read through the index and pair files matching up sequences to barcodes
    missing_barcodes=0
    total_barcodes=0
    read_counts_by_sample={}
    new_files=set()
    if args.input_index:
        all_index_lines = read_fastq_file(args.input_index)
    else:
        all_index_lines = read_fastq_file(args.input_read1)

    for index_lines in all_index_lines:
        total_barcodes+=1
       
        if args.input_index: 
            sequence=index_lines[1].rstrip()
        else:
            sequence=index_lines[0].rstrip().split(":")[-1]
        
        # reverse complement
        if args.reverse_complement:
            sequence=sequence.translate(rev_cmp)[::-1]
        
        try:
            sample_id=barcodes[sequence]
        except KeyError:
            missing_barcodes+=1
            sample_id="unknown"
            
        # check that read1/2 have the same sequence id
        read1=next(read1_lines)
        read1_id = read1[0].split(" ")[0]
        read2=next(read2_lines)
        read2_id = read2[0].split(" ")[0]

        # get sequence id depending on if index is provided
        if args.input_index:
            sequence_id=index_lines[0].split(" ")[0]
        else:
            sequence_id=read1_id
        
        if not ( (read1_id == read2_id) and (read2_id == sequence_id) ):
            sys.exit("Reads are not ordered: " + read1_id)
            
        # increase read counts by sample
        read_counts_by_sample[sample_id]=read_counts_by_sample.get(sample_id,0)+1
        
        # write pairs to new files
        new_read1_file=os.path.join(args.output,sample_id+"_R1.fastq")
        new_read2_file=os.path.join(args.output,sample_id+"_R2.fastq")
        
        # add to the list of new files
        new_files.add(new_read1_file)
        new_files.add(new_read2_file)
        
        # open each file in append mode
        # can't keep all files open as this might be too many open files
        # depending on the total number of samples
        with open(new_read1_file, "a") as fh:
            fh.write("".join(read1))
            
        with open(new_read2_file, "a") as fh:
            fh.write("".join(read2))
        
    # write the read counts to the file
    with open(os.path.join(args.output,"read_counts.txt"),"w") as fh:
        fh.write("# sample\ttotal_reads\n")
        for sample, count in read_counts_by_sample.items():
            fh.write("{}\t{}\n".format(sample,count))
        
    print("Total barcodes: {}".format(total_barcodes))
    print("Unknown barcodes: {} {}%".format(missing_barcodes,missing_barcodes/(total_barcodes*1.0)*100))
    print("Files written: "+"\n".join(sorted(list(new_files))))

if __name__ == "__main__":
    main()
